# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import TYPE_CHECKING, Tuple, Union

import jax
import jax.numpy as jnp
from jax.sharding import Mesh
from vllm.model_executor.layers.fused_moe import FusedMoE

from tpu_inference import envs
from tpu_inference.kernels.fused_moe.v1.kernel import fused_ep_moe
from tpu_inference.layers.common.fused_moe_gmm import fused_moe_func
from tpu_inference.logger import init_logger

if TYPE_CHECKING:
    from tpu_inference.layers.common.process_weights.moe_weights import (
        FusedMoEWeights, UnfusedMoEWeights)
    from tpu_inference.layers.jax.moe.moe import JaxMoE
else:
    FusedMoEWeights = None
    UnfusedMoEWeights = None
    JaxMoE = None

logger = init_logger(__name__)


class MoEBackend(Enum):
    # FUSED_MOE is using the Fused MoE kernel found in tpu-inference/tpu_inference/kernels/fused_moe/v1/kernel.py
    # and is production ready.
    # NOTE: for FusedMOE on the JAX/Flax path, we expect the MoE quant method (e.g. UnquantizedFusedMoEMethod) to
    # create a new `kernel_gating_upproj_E2DF` param that replaces `kernel_gating_EDF` and `kernel_up_proj_EDF`
    # and thus runs the forward pass on the fused weight.  `kernel_down_proj_EFD` is unchanged.
    FUSED_MOE = "fused_moe"

    # GMM_EP is using the GMM kernel found in tpu-inference/tpu_inference/layers/common/fused_moe_gmm.py
    # and is `expert_sharded_gmm` the GMM calls.  Production ready.
    # NOTE: for GMM_EP JAX/Flax path, we expect the MoE quant method (e.g. UnquantizedFusedMoEMethod) to
    # create a new `kernel_gating_upproj_EDF` param that replaces `kernel_gating_EDF` and `kernel_up_proj_EDF`
    # and thus runs the forward pass on the fused weight.  `kernel_down_proj_EFD` is unchanged.
    GMM_EP = "gmm_ep"

    # Same as GMM_EP but uses `tensor_sharded_gmm_*` for the GMM calls and is production ready
    GMM_TP = "gmm_tp"

    # DENSE_MAT uses a simple dense matmul for the MoE backend,  is intended for testing, and is
    # only used in the JAX path for now
    # NOTE: for DENSE_MAT in the JAX/Flax path, there are no changes for weights for the unfused backends (DENSE_MAT and MEGABLOX_GMM).
    # That is, `kernel_gating_EDF`, `kernel_up_proj_EDF`, and `kernel_down_proj_EFD` are unchanged.
    DENSE_MAT = "dense_mat"
    # Also only used in the JAX path for now
    MEGABLX_GMM = "megablox_gmm"

    @classmethod
    def fused_moe_backends(cls):
        """Returns those backends that use fused weights"""
        return {cls.FUSED_MOE, cls.GMM_EP, cls.GMM_TP}


def moe_apply(
    layer: Union[FusedMoE, JaxMoE],
    x: jax.Array,
    gating_output: Union[jax.Array, Tuple[jax.Array, jax.Array]],
    weights: Union[FusedMoEWeights, UnfusedMoEWeights],
    moe_backend: MoEBackend,
    mesh: Mesh,
    extra_backend_kwargs: dict,
) -> jax.Array:

    with jax.named_scope(layer._get_name()):
        activation = layer.activation if isinstance(
            layer.activation, str) else layer.activation.value
        match moe_backend:
            case MoEBackend.FUSED_MOE:
                subc_quant_w1_sz = None
                subc_quant_w2_sz = None
                if weights.w13_weight_scale is not None and weights.w2_weight_scale is not None:
                    padded_hidden_size = weights.w13_weight.shape[-2]
                    # NB: w13_weight_scale: (num_experts, 2, hidden_size // subc_quant_w1_sz, 1, intermediate_size)
                    assert padded_hidden_size % weights.w13_weight_scale.shape[
                        2] == 0
                    subc_quant_w1_sz = padded_hidden_size // weights.w13_weight_scale.shape[
                        2]
                    intermediate_size = weights.w13_weight.shape[-1]
                    # NB: w2_weight_scale: (num_experts, intermediate_size // subc_quant_w2_sz, 1, hidden_size)
                    assert intermediate_size % weights.w2_weight_scale.shape[
                        1] == 0
                    subc_quant_w2_sz = intermediate_size // weights.w2_weight_scale.shape[
                        1]

                actual_hidden_size = x.shape[-1]
                padding_size = weights.w13_weight.shape[-2] - actual_hidden_size
                x = jnp.pad(x, ((0, 0), (0, padding_size)))
                output = fused_ep_moe(
                    mesh=mesh,
                    tokens=x,
                    w1=weights.w13_weight,
                    w2=weights.w2_weight,
                    gating_output=gating_output,
                    top_k=layer.top_k,
                    renormalize_topk_logits=layer.renormalize,
                    act_fn=activation,
                    scoring_fn=layer.scoring_func,
                    subc_quant_w1_sz=subc_quant_w1_sz,
                    subc_quant_w2_sz=subc_quant_w2_sz,
                    w1_scale=weights.w13_weight_scale,
                    w2_scale=weights.w2_weight_scale,
                    b1=weights.w13_bias,
                    b2=weights.w2_bias,
                    **extra_backend_kwargs,
                )[:, :actual_hidden_size]
            case MoEBackend.GMM_EP | MoEBackend.GMM_TP:
                output = fused_moe_func(
                    hidden_states=x,
                    w1=weights.w13_weight,
                    w2=weights.w2_weight,
                    w1_scale=weights.w13_weight_scale,
                    w2_scale=weights.w2_weight_scale,
                    w1_bias=weights.w13_bias,
                    w2_bias=weights.w2_bias,
                    gating_output=gating_output,
                    topk=layer.top_k,
                    renormalize=layer.renormalize,
                    mesh=mesh,
                    use_ep=layer.use_ep,
                    activation=activation,
                    scoring_fn=layer.scoring_func,
                    sc_kernel_threshold=envs.SC_KERNEL_THRESHOLD,
                    sc_kernel_col_chunk_size=envs.SC_KERNEL_COL_CHUNK_SIZE,
                )
            case MoEBackend.DENSE_MAT:
                # NOTE: circular import avoidance
                from tpu_inference.layers.jax.moe.dense_moe import \
                    dense_moe_func
                assert isinstance(
                    gating_output,
                    tuple), "Expected the gating output to be a tuple"
                assert len(
                    gating_output
                ) == 2, "Expected the gating output to be have 2 entries: weights and indices"
                return dense_moe_func(
                    weights=weights,
                    x_TD=x,
                    gating_output=gating_output,
                    cast_dtype=layer.dtype,
                    num_local_experts=layer.num_local_experts,
                    apply_expert_weight_before_computation=layer.
                    apply_expert_weight_before_computation,
                    activation_ffw_ted=layer.activation_ffw_ted,
                    activation_ffw_td=layer.activation_ffw_td,
                    hidden_act=layer.hidden_act,
                    mesh=mesh)

            case MoEBackend.MEGABLX_GMM:
                # NOTE: circular import avoidance
                from tpu_inference.layers.jax.moe.sparse_moe import \
                    sparse_moe_func

                return sparse_moe_func(weights=weights,
                                       x_TD=x,
                                       gating_output=gating_output,
                                       layer=layer,
                                       mesh=mesh)

        return output



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import jax
import jax.numpy as jnp
import torch
from compressed_tensors.quantization import (QuantizationArgs,
                                             QuantizationStrategy)
from jax.sharding import NamedSharding, PartitionSpec
from torch.nn.parameter import Parameter
from torchax.interop import jax_view, torch_view
from vllm.model_executor.layers.quantization.compressed_tensors.schemes.compressed_tensors_w8a8_fp8 import \
    CompressedTensorsW8A8Fp8
from vllm.model_executor.layers.quantization.utils.fp8_utils import \
    W8A8BlockFp8LinearOp
from vllm.model_executor.layers.quantization.utils.quant_utils import \
    GroupShape

from tpu_inference.layers.common.linear import sharded_quantized_matmul
from tpu_inference.layers.common.process_weights.linear_weights import (
    LinearWeights, process_linear_weights, shard_linear_weights,
    to_parameter_list)
from tpu_inference.layers.common.quantization import (dequantize_tensor,
                                                      quantize_tensor)
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation
from tpu_inference.layers.vllm.quantization.configs import \
    VllmQuantLinearConfig
from tpu_inference.logger import init_logger
from tpu_inference.utils import t2j

P = PartitionSpec

logger = init_logger(__name__)


class VllmCompressedTensorsW8A8Fp8(CompressedTensorsW8A8Fp8):

    def __init__(
        self,
        weight_quant: QuantizationArgs,
        is_static_input_scheme: bool,
        linear_config: VllmQuantLinearConfig,
    ):
        self.weight_quant = weight_quant
        self.strategy = weight_quant.strategy
        self.out_dtype = torch.get_default_dtype()
        self.is_static_input_scheme = is_static_input_scheme
        self.weight_block_size = self.weight_quant.block_structure

        if self.weight_block_size is not None:
            assert not self.is_static_input_scheme
            self.act_q_group_shape = GroupShape(1, self.weight_block_size[0])
            self.w8a8_block_fp8_linear = W8A8BlockFp8LinearOp(
                weight_group_shape=GroupShape(*self.weight_block_size),
                act_quant_group_shape=self.act_q_group_shape,
                cutlass_block_fp8_supported=self.cutlass_block_fp8_supported,
                use_aiter_and_is_supported=self.use_aiter_and_is_supported,
            )

        self.linear_config = linear_config

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        weight = t2j(layer.weight, use_dlpack=False)
        delattr(layer, "weight")
        weight_scale = t2j(layer.weight_scale, use_dlpack=False)
        delattr(layer, "weight_scale")

        if layer.bias is not None and not layer.skip_bias_add:
            if layer.return_bias:
                logger.warning_once("Bias might return incorrect value.")
            bias = t2j(layer.bias, use_dlpack=False)
            delattr(layer, "bias")
        else:
            bias = None

        per_tensor = self.strategy == QuantizationStrategy.TENSOR

        @jax.jit
        def process_fp8_linear_weights(
            weight: jax.Array,
            weight_scale: jax.Array,
            bias: jax.Array | None,
        ) -> LinearWeights:
            if per_tensor:
                weights = []
                start = 0
                # Multiple weights may have been concatenated. Loop through
                # each weight and perform dequantization.
                for i, output_size in enumerate(
                        self.linear_config.output_sizes):
                    end = start + output_size
                    weights.append(
                        dequantize_tensor(weight[start:end], weight_scale[i]))
                    start = end
                weight = jnp.concat(weights, axis=0)
                # Requantize into per-tensor.
                weight, weight_scale = quantize_tensor(jnp.float8_e4m3fn,
                                                       weight, None)
            else:
                weight_scale = jnp.squeeze(weight_scale, -1)

            return process_linear_weights(
                LinearWeights(
                    weight=weight,
                    weight_scale=weight_scale,
                    zero_point=None,
                    bias=bias,
                ),
                fused=self.linear_config.fuse_matmuls,
                output_sizes=self.linear_config.output_sizes,
                reorder_size=self.linear_config.n_shards,
                per_tensor=per_tensor,
            )

        weights = process_fp8_linear_weights(weight, weight_scale, bias)
        weights = torch_view(
            shard_linear_weights(
                weights,
                mesh=self.linear_config.mesh,
                weight_p_spec=self.linear_config.weight_sharding,
                bias_p_spec=self.linear_config.bias_sharding,
                per_tensor=per_tensor,
            ))

        if self.linear_config.fuse_matmuls:
            layer.weight = Parameter(weights.weight, requires_grad=False)
            layer.weight_scale = Parameter(weights.weight_scale,
                                           requires_grad=False)
            if bias is not None:
                layer.bias = Parameter(weights.bias, requires_grad=False)
        else:
            layer.weight = to_parameter_list(weights.weight)
            layer.weight_scale = to_parameter_list(weights.weight_scale)
            if bias is not None:
                layer.bias = to_parameter_list(weights.bias)

        if self.is_static_input_scheme:
            # In static quant, all input_scales share the same value.
            assert layer.input_scale.min() == layer.input_scale.max()
            input_scale_first = layer.input_scale[0]

            input_scale = jax.device_put(
                t2j(input_scale_first, use_dlpack=False),
                NamedSharding(self.linear_config.mesh, P()))
            input_scale = torch.nn.Parameter(torch_view(input_scale),
                                             requires_grad=False)
            delattr(layer, "input_scale")
            layer.input_scale = input_scale

    def apply_weights(self, layer: torch.nn.Module, x: torch.Tensor,
                      bias: Optional[torch.Tensor]) -> torch.Tensor:
        with jax.named_scope(layer._get_name()):
            if self.linear_config.fuse_matmuls:
                return self._apply_fused(layer, x, bias)
            else:
                return self._apply_split(layer, x, bias)

    def _apply_fused(self, layer: torch.nn.Module, x: torch.Tensor,
                     bias: Optional[torch.Tensor]) -> torch.Tensor:
        x_jax = jax_view(x)
        weight_jax = jax_view(layer.weight)
        weight_scale_jax = jax_view(layer.weight_scale)

        if self.is_static_input_scheme:
            # TODO(kyuyeunk): Add kernel support for static quant
            input_scale = jax_view(layer.input_scale)
            dtype_info = jnp.finfo(weight_jax.dtype)
            maxval = float(dtype_info.max)
            minval = float(dtype_info.min)
            x_q = jnp.clip(x_jax / input_scale.astype(x_jax.dtype), minval,
                           maxval).astype(weight_jax.dtype)

            outs = jax.lax.dot_general(
                x_q,
                weight_jax,
                (((1, ), (1, )), ((), ())),
                preferred_element_type=jnp.float32,
            )
            outs *= weight_scale_jax
            outs = outs.astype(x_jax.dtype)
        else:
            outs = sharded_quantized_matmul(x_jax,
                                            weight_jax,
                                            weight_scale_jax,
                                            self.linear_config.weight_sharding,
                                            mesh=self.linear_config.mesh)

        if bias is not None and not layer.skip_bias_add:
            outs += jax_view(bias)
        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        return torch_view(jnp.concatenate(outs, axis=-1))

    def _apply_split(self, layer: torch.nn.Module, x: torch.Tensor,
                     bias: Optional[torch.Tensor]) -> torch.Tensor:
        assert isinstance(layer.weight, torch.nn.ParameterList)

        x_jax = jax_view(x)
        outs = []
        for i, (weight, weight_scale) in enumerate(
                zip(layer.weight, layer.weight_scale)):
            weight_jax = jax_view(weight)
            weight_scale_jax = jax_view(weight_scale)

            if self.is_static_input_scheme:
                # TODO(kyuyeunk): Add kernel support for static quant
                input_scale = jax_view(layer.input_scale)
                dtype_info = jnp.finfo(weight_jax.dtype)
                maxval = float(dtype_info.max)
                minval = float(dtype_info.min)
                x_q = jnp.clip(x_jax / input_scale.astype(x_jax.dtype), minval,
                               maxval).astype(weight_jax.dtype)

                out = jax.lax.dot_general(
                    x_q,
                    weight_jax,
                    (((1, ), (1, )), ((), ())),
                    preferred_element_type=jnp.float32,
                )
                # TODO(kyuyeunk): Investigate performance gain from merging scales.
                # out *= weight_scale_jax
                out *= weight_scale_jax * input_scale
                out = out.astype(x_jax.dtype)
            else:
                out = sharded_quantized_matmul(
                    x_jax,
                    weight_jax,
                    weight_scale_jax,
                    self.linear_config.weight_sharding,
                    mesh=self.linear_config.mesh)

            if bias is not None and not layer.skip_bias_add:
                out += jax_view(bias[i])
            outs.append(out)
        return torch_view(jnp.concatenate(outs, axis=-1))



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Optional, Sequence

import jax
from jax import numpy as jnp
from jax.sharding import Mesh

from tpu_inference.layers.common.linear import sharded_quantized_matmul
from tpu_inference.layers.common.process_weights.linear_weights import (
    LinearWeights, process_linear_weights)
from tpu_inference.layers.common.quantization import (dequantize_tensor,
                                                      quantize_tensor)
from tpu_inference.layers.common.quantization.configs import QuantLinearConfig
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation


class Fp8LinearMethod:
    """Implements the forward method for fp8 linear layers.

    This class will be shared in both vLLM and jax path.
    """

    def __init__(self, linear_config: QuantLinearConfig):
        self.linear_config = linear_config

    def _apply_fused(self, x: jax.Array, weight_jax: jax.Array,
                     weight_scale_jax: jax.Array,
                     bias: Optional[jax.Array]) -> jax.Array:
        outs = sharded_quantized_matmul(x,
                                        weight_jax,
                                        weight_scale_jax,
                                        self.linear_config.weight_sharding,
                                        mesh=self.linear_config.mesh)

        if bias is not None:
            outs += bias
        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        return jnp.concatenate(outs, axis=-1)

    def _apply_split(self,
                     x: jax.Array,
                     weight_and_scale: Sequence[tuple[jax.Array, jax.Array]],
                     bias: Optional[Sequence[jax.Array]] = None,
                     mesh: Optional[Mesh] = None) -> jax.Array:

        outs = []
        for i, (weight, weight_scale) in enumerate(weight_and_scale):

            out = sharded_quantized_matmul(x,
                                           weight,
                                           weight_scale,
                                           self.linear_config.weight_sharding,
                                           mesh=mesh)

            if bias is not None:
                out += bias[i]
            outs.append(out)
        return jnp.concatenate(outs, axis=-1)


@jax.jit(static_argnames=(
    'weight_block_size',
    'requant_block_size',
    'output_sizes',
    'requant_weight_dtype',
    'fuse_matmuls',
    'n_shards',
))
def process_blockwise_fp8_linear_weights(
    weight: jax.Array,
    weight_scale: jax.Array,
    *,
    bias: jax.Array | None,
    weight_block_size: Sequence[int],
    requant_block_size,
    output_sizes,
    requant_weight_dtype,
    fuse_matmuls,
    n_shards,
) -> LinearWeights:
    weights = []
    weight_scales = []
    original_block_size = weight_block_size[0]
    start = 0
    for output_size in output_sizes:
        end = start + output_size

        weight_slice = weight[start:end]
        weight_scale_slice = weight_scale[start // original_block_size:math.
                                          ceil(end / original_block_size)]
        dequantized_weight = dequantize_tensor(
            weight_slice,
            weight_scale_slice,
            (0, 1),
            block_size=weight_block_size,
        )
        weight_slice, weight_scale_slice = quantize_tensor(
            requant_weight_dtype,
            dequantized_weight,
            block_size=requant_block_size)

        weights.append(weight_slice)
        weight_scales.append(weight_scale_slice)

        start = end

    weight = jnp.concat(weights, axis=0)
    weight_scale = jnp.concat(weight_scales, axis=0)

    return process_linear_weights(
        LinearWeights(
            weight=weight,
            weight_scale=weight_scale,
            zero_point=None,
            bias=bias,
        ),
        fused=fuse_matmuls,
        output_sizes=output_sizes,
        reorder_size=n_shards,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import regex as re
import torch
from vllm.config import ModelConfig, VllmConfig
from vllm.config.load import LoadConfig
from vllm.model_executor.model_loader import register_model_loader
from vllm.model_executor.model_loader.default_loader import DefaultModelLoader
from vllm.model_executor.model_loader.runai_streamer_loader import \
    RunaiModelStreamerLoader
from vllm.model_executor.model_loader.utils import (
    initialize_model, process_weights_after_loading)
from vllm.utils.torch_utils import set_default_torch_dtype

from tpu_inference.layers.vllm.quantization.base import VllmQuantizationMethod


def attach_incremental_weight_loader(model: torch.nn.Module) -> None:
    """
    Traverses the model and overrides the weight_loader of each parameter to support incremental loading.
    This allows processing and sharding of weights after all weights for a module have been loaded.
    """

    def create_weight_loader(layer, original_loader, layer_name, param_name):

        def weight_loader_wrapper(param: torch.nn.Parameter,
                                  loaded_weight: torch.Tensor, *args,
                                  **kwargs):
            # Loading the weight
            res = original_loader(param, loaded_weight, *args, **kwargs)

            # Processing and sharding
            # For now, only handle unquantized linear and moe layers.
            quant_method = layer.quant_method
            if isinstance(quant_method, VllmQuantizationMethod):
                quant_method.maybe_process_weights(layer, param_name, args,
                                                   kwargs)

            return res

        return weight_loader_wrapper

    for name, module in model.named_modules():
        # Weight loader will be invoked multiple times for module. In order to determine when all the weights are loaded,
        # we need to keep track of the loaded weights for each module.
        module._loaded_weights = set()
        for param_name, param in module.named_parameters(recurse=False):
            # Omit parameters that do not have a weight_loader
            original_loader = getattr(param, "weight_loader", None)
            if original_loader is None:
                continue
            setattr(
                param, "weight_loader",
                create_weight_loader(module, original_loader, name,
                                     param_name))


@register_model_loader("tpu_streaming_loader")
class IncrementalModelLoader(DefaultModelLoader):
    """
    Model loader that supports incremental weight loading and sharding.

    This loader is needed to inject the `attach_incremental_weight_loader` logic
    before the actual weight loading begins. This allows us to wrap the
    parameter weight loaders so that weights are sharded to TPU and freed from
    CPU memory as soon as a layer is fully loaded, rather than waiting for the
    entire model to be loaded into CPU memory first.
    """

    def __init__(self, load_config: LoadConfig):
        load_config.load_format = "auto"
        super().__init__(load_config)

    def load_model(self,
                   vllm_config: VllmConfig,
                   model_config: ModelConfig,
                   prefix: str = "") -> torch.nn.Module:
        """Load a model with the given configurations."""
        device_config = vllm_config.device_config
        load_config = vllm_config.load_config
        load_device = (device_config.device
                       if load_config.device is None else load_config.device)
        target_device = torch.device(load_device)
        with set_default_torch_dtype(model_config.dtype):
            with target_device:
                model = initialize_model(vllm_config=vllm_config,
                                         model_config=model_config)
            # Override weight loader logic of each parameter to support incremental loading.
            attach_incremental_weight_loader(model)
            # Quantization does not happen in `load_weights` but after it
            self.load_weights(model, model_config)
            process_weights_after_loading(model, model_config, target_device)

        return model.eval()


@register_model_loader("runai_streamer")
class RunaiIncrementalModelLoader(RunaiModelStreamerLoader):
    """Model loader that supports both RunAI streaming and incremental weight sharding."""

    def __init__(self, load_config: LoadConfig):
        super().__init__(load_config)

    def _prepare_weights(self, model_name_or_path: str,
                         revision: str | None) -> list[str]:
        hf_weights_files = super()._prepare_weights(model_name_or_path,
                                                    revision)
        hf_weights_files.sort(key=lambda f: [
            int(s) if s.isdigit() else s
            for s in re.split(r"(\d+)", os.path.basename(f))
        ])
        return hf_weights_files

    def load_model(self,
                   vllm_config: VllmConfig,
                   model_config: ModelConfig,
                   prefix: str = "") -> torch.nn.Module:
        """Load a model with the given configurations."""
        device_config = vllm_config.device_config
        load_config = vllm_config.load_config
        load_device = (device_config.device
                       if load_config.device is None else load_config.device)
        target_device = torch.device(load_device)
        with set_default_torch_dtype(model_config.dtype):
            with target_device:
                model = initialize_model(vllm_config=vllm_config,
                                         model_config=model_config)
            # Override weight loader logic of each parameter to support incremental loading.
            attach_incremental_weight_loader(model)
            # Quantization does not happen in `load_weights` but after it
            self.load_weights(model, model_config)
            process_weights_after_loading(model, model_config, target_device)

        return model.eval()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax.numpy as jnp
import torch
import torchax
from jax.sharding import NamedSharding
from jax.sharding import PartitionSpec as P
from torch.nn import Parameter
from torchax.interop import jax_view, torch_view
from vllm.config import CacheConfig
from vllm.model_executor.layers.attention.attention import \
    get_attention_context
from vllm.model_executor.layers.attention.mla_attention import MLAAttention
from vllm.model_executor.layers.linear import ColumnParallelLinear
from vllm.model_executor.layers.mla import (MLAModules,
                                            MultiHeadLatentAttentionWrapper)
from vllm.model_executor.layers.quantization import QuantizationConfig
from vllm.v1.attention.backend import AttentionType

from tpu_inference import utils
from tpu_inference.layers.common.quantization import quantize_tensor
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.layers.common.utils import general_device_put
from tpu_inference.models.vllm.vllm_model_wrapper_context import \
    get_vllm_model_wrapper_context


class VllmMLAAttention(MLAAttention):

    def __init__(
        self,
        num_heads: int,
        scale: float,
        qk_nope_head_dim: int,
        qk_rope_head_dim: int,
        v_head_dim: int,
        q_lora_rank: int | None,
        kv_lora_rank: int,
        kv_b_proj: ColumnParallelLinear,
        cache_config: CacheConfig | None = None,
        quant_config: QuantizationConfig | None = None,
        prefix: str = "",
        use_sparse: bool = False,
        indexer: object | None = None,
        **extra_impl_args,
    ):
        torch.nn.Module.__init__(self)
        super().__init__(num_heads, scale, qk_nope_head_dim, qk_rope_head_dim,
                         v_head_dim, q_lora_rank, kv_lora_rank, kv_b_proj,
                         cache_config, quant_config, prefix, use_sparse,
                         indexer, **extra_impl_args)

        # For compatibility reasons.
        self.kv_sharing_target_layer_name = None
        self.attn_type = AttentionType.DECODER
        self.sliding_window = None

        self.kv_cache_quantized_dtype = None
        if self.kv_cache_dtype != "auto":
            self.kv_cache_quantized_dtype = utils.to_jax_dtype(
                self.kv_cache_dtype)

    def process_weights_after_loading(self, act_dtype: torch.dtype):
        with torchax.default_env():
            super().process_weights_after_loading(act_dtype)

            # NOTE: vLLM dequantizes kv_b_proj weights which causes more memory
            # usage than expected.

            # quantize W_UK_T, W_UV back to cache type and transfer
            # `W_UK_T`, `W_UV` to TPUs
            mesh = self.kv_b_proj.quant_method.linear_config.mesh
            sharding = NamedSharding(mesh, P(ShardingAxisName.ATTN_HEAD, ))
            self.W_UK_T, self.W_UK_T_scale = quantize_tensor(
                self.kv_cache_quantized_dtype, jax_view(self.W_UK_T), axis=1)
            self.W_UK_T = torch_view(general_device_put(self.W_UK_T, sharding))
            self.W_UK_T_scale = torch_view(
                general_device_put(jnp.expand_dims(self.W_UK_T_scale, 0),
                                   sharding))

            self.W_UV, self.W_UV_scale = quantize_tensor(
                self.kv_cache_quantized_dtype, jax_view(self.W_UV), axis=1)
            self.W_UV = torch_view(general_device_put(self.W_UV, sharding))
            self.W_UV_scale = torch_view(
                general_device_put(jnp.expand_dims(self.W_UV_scale, 0),
                                   sharding))

            self.W_UK_T = Parameter(self.W_UK_T, requires_grad=False)
            self.W_UK_T_scale = Parameter(self.W_UK_T_scale,
                                          requires_grad=False)
            self.W_UV = Parameter(self.W_UV, requires_grad=False)
            self.W_UV_scale = Parameter(self.W_UV_scale, requires_grad=False)

            # Delete kv_b_proj_params as the dequantized weights are now stored
            # in self.W_UK_T and self.W_UV.
            kv_b_proj_params = dict(self.kv_b_proj.named_parameters())
            for key in kv_b_proj_params.keys():
                delattr(self.kv_b_proj, key)

    def forward(self, q: torch.Tensor, kv_c_normed: torch.Tensor,
                k_pe: torch.Tensor, **kwargs) -> torch.Tensor:
        if self.calculate_kv_scales:
            torch.ops.vllm.maybe_calc_kv_scales(q, kv_c_normed, k_pe,
                                                self.layer_name)

        # Get the KV cache
        vllm_model_wrapper_context = get_vllm_model_wrapper_context()
        kv_cache_index = vllm_model_wrapper_context.layer_name_to_kvcache_index[
            self.layer_name]
        kv_cache = vllm_model_wrapper_context.kv_caches[kv_cache_index]

        # Get the mesh
        mesh = vllm_model_wrapper_context.mesh

        # Get the attention metadata
        attn_metadata, _, _, _ = get_attention_context(self.layer_name)

        # Run the fundamental MLA forward pass from the impl
        outputs, new_kv_cache = self.impl.forward(q, kv_c_normed, k_pe,
                                                  kv_cache, attn_metadata,
                                                  mesh, self, **kwargs)

        # Update KV cache
        vllm_model_wrapper_context.kv_caches[kv_cache_index] = new_kv_cache

        return torch_view(outputs)


@MultiHeadLatentAttentionWrapper.register_oot
class VllmMultiHeadLatentAttentionWrapper(MultiHeadLatentAttentionWrapper):

    def __init__(
        self,
        hidden_size: int,
        num_heads: int,
        scale: float,
        qk_nope_head_dim: int,
        qk_rope_head_dim: int,
        v_head_dim: int,
        q_lora_rank: int | None,
        kv_lora_rank: int,
        mla_modules: MLAModules,
        cache_config: CacheConfig | None = None,
        quant_config: QuantizationConfig | None = None,
        prefix: str = "",
    ) -> None:
        torch.nn.Module.__init__(self)

        self.hidden_size = hidden_size
        self.qk_nope_head_dim = qk_nope_head_dim
        self.qk_rope_head_dim = qk_rope_head_dim
        self.qk_head_dim = qk_nope_head_dim + qk_rope_head_dim
        self.v_head_dim = v_head_dim
        self.q_lora_rank = q_lora_rank
        self.kv_lora_rank = kv_lora_rank
        self.num_heads = num_heads
        self.fused_qkv_a_proj = mla_modules.fused_qkv_a_proj
        self.kv_a_proj_with_mqa = mla_modules.kv_a_proj_with_mqa
        self.q_a_layernorm = mla_modules.q_a_layernorm
        self.q_b_proj = mla_modules.q_b_proj
        self.q_proj = mla_modules.q_proj
        self.kv_a_layernorm = mla_modules.kv_a_layernorm
        self.kv_b_proj = mla_modules.kv_b_proj
        self.rotary_emb = mla_modules.rotary_emb
        self.o_proj = mla_modules.o_proj
        self.indexer = mla_modules.indexer
        self.indexer_rope_emb = mla_modules.indexer_rotary_emb
        self.is_sparse = mla_modules.is_sparse

        if self.indexer is not None:
            assert hasattr(self.indexer, "topk_tokens")
            self.topk_tokens = self.indexer.topk_tokens
            self.topk_indices_buffer = mla_modules.topk_indices_buffer

        self.mla_attn = VllmMLAAttention(
            num_heads=self.num_heads,
            scale=scale,
            qk_nope_head_dim=self.qk_nope_head_dim,
            qk_rope_head_dim=self.qk_rope_head_dim,
            v_head_dim=self.v_head_dim,
            q_lora_rank=self.q_lora_rank,
            kv_lora_rank=self.kv_lora_rank,
            cache_config=cache_config,
            quant_config=quant_config,
            prefix=f"{prefix}.attn",
            kv_b_proj=self.kv_b_proj,
            use_sparse=self.is_sparse,
            indexer=self.indexer,
        )

        self.prefix = prefix



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
from vllm.config import (CacheConfig, ModelConfig, ParallelConfig,
                         SchedulerConfig, SpeculativeConfig, VllmConfig)
from vllm.model_executor.layers.rotary_embedding import MRotaryEmbedding
from vllm.multimodal.inputs import (MultiModalBatchedField,
                                    MultiModalFeatureSpec, MultiModalFieldElem,
                                    MultiModalKwargsItem, PlaceholderRange)
from vllm.sampling_params import SamplingType
from vllm.v1.core.sched.output import SchedulerOutput as VllmSchedulerOutput

from tpu_inference.runner.input_batch import CachedRequestState
from tpu_inference.runner.tpu_runner import TPUModelRunner


class TestMultiModalManager:

    def setup_method(self):
        # Mock JAX dependencies
        self.mock_devices = [MagicMock(coords=i) for i in range(1)]
        device_array = np.array(jax.devices()[:1]).reshape(1, 1, 1, 1)
        self.mock_mesh = jax.make_mesh(device_array.shape,
                                       ('data', 'attn_dp', 'expert', 'model'))
        self.mock_rng_key = MagicMock()

        with patch('jax.devices', return_value=self.mock_devices), \
             patch('jax.make_mesh', return_value=self.mock_mesh), \
             patch('jax.random.key', return_value=self.mock_rng_key), \
             patch('tpu_inference.runner.tpu_runner.get_model', return_value=MagicMock()), \
             patch('tpu_inference.runner.tpu_runner.make_optimized_mesh', return_value=self.mock_mesh):

            model_config = ModelConfig(tokenizer_mode="auto",
                                       trust_remote_code=False,
                                       seed=0,
                                       dtype='bfloat16')
            cache_config = CacheConfig(
                block_size=16,
                gpu_memory_utilization=0.9,
                cache_dtype="auto",
            )
            scheduler_config = SchedulerConfig(max_num_seqs=16,
                                               max_model_len=1024,
                                               is_encoder_decoder=False)
            parallel_config = ParallelConfig(
                pipeline_parallel_size=1,
                tensor_parallel_size=1,
            )
            speculative_config = SpeculativeConfig(
                model='ngram',
                num_speculative_tokens=5,
                prompt_lookup_max=4,
            )
            vllm_config = VllmConfig(
                model_config=model_config,
                cache_config=cache_config,
                scheduler_config=scheduler_config,
                parallel_config=parallel_config,
                speculative_config=speculative_config,
                observability_config={},
                additional_config={},
            )

            self.runner = TPUModelRunner(vllm_config,
                                         devices=self.mock_devices)

    def test_execute_mm_encoder_single_image(self):
        import torch
        """Tests _execute_mm_encoder with a single request and a single image."""
        # 1. ===== Setup =====
        self.runner.is_multimodal_model = True
        self.mock_get_mm_embed_fn = MagicMock()
        self.runner.embed_multimodal_fn = self.mock_get_mm_embed_fn

        self.runner.state = MagicMock()
        # Mock scheduler output
        mock_scheduler_output = MagicMock(spec=VllmSchedulerOutput)
        mock_scheduler_output.scheduled_encoder_inputs = {"req-1": [0]}

        # Mock request state
        dummy_pixel_values = torch.randn(3, 224, 224, dtype=torch.bfloat16)
        dummy_grid_thw = torch.tensor([[1, 1, 1]], dtype=torch.int64)
        mm_item = MultiModalKwargsItem({
            "pixel_values":
            MultiModalFieldElem(dummy_pixel_values, MultiModalBatchedField()),
            "image_grid_thw":
            MultiModalFieldElem(dummy_grid_thw, MultiModalBatchedField())
        })

        req_state = CachedRequestState(
            req_id="req-1",
            prompt_token_ids=[1, 2, 3],
            output_token_ids=[],
            sampling_params=MagicMock(),
            block_ids=(),
            num_computed_tokens=0,
            mm_features=[
                MultiModalFeatureSpec(data=mm_item,
                                      identifier="req-1",
                                      modality="image",
                                      mm_position=PlaceholderRange(offset=0,
                                                                   length=1))
            ],
            lora_request=None,
            pooling_params=None,
            generator=None,
        )
        self.runner.requests = {"req-1": req_state}

        # Mock the return value of the multimodal encoder
        dummy_embedding = jnp.ones((10, 128), dtype=jnp.bfloat16)
        self.mock_get_mm_embed_fn.return_value = (dummy_embedding, )

        # 2. ===== Act =====
        self.runner.mm_manager.execute_mm_encoder(mock_scheduler_output)

        # 3. ===== Assert =====
        # Check if encoder_cache is populated correctly
        assert "req-1" in self.runner.encoder_cache
        cached_embedding = self.runner.encoder_cache["req-1"]
        np.testing.assert_array_equal(np.asarray(cached_embedding),
                                      np.asarray(dummy_embedding))

        # Check if embed_multimodal_fn was called with correct args
        self.mock_get_mm_embed_fn.assert_called_once()
        call_args = self.mock_get_mm_embed_fn.call_args

        # Positional args: (state, image_grid_thw)
        state_arg, grid_arg = call_args.args
        # Keyword args: **batched_mm_inputs
        kwargs_arg = call_args.kwargs

        assert state_arg == self.runner.state
        assert grid_arg == ((1, 1, 1), )
        assert "pixel_values" in kwargs_arg

        # Verify the pixel values tensor passed to the mock
        passed_pixel_values = kwargs_arg['pixel_values']
        assert isinstance(passed_pixel_values, np.ndarray)
        assert passed_pixel_values.dtype == jnp.bfloat16

        # Convert torch tensor for comparison
        expected_pixel_values = dummy_pixel_values.unsqueeze(0).to(
            torch.float32).numpy().astype(jnp.bfloat16)
        np.testing.assert_array_equal(np.asarray(passed_pixel_values),
                                      expected_pixel_values)

    def test_execute_mm_encoder_multiple_images(self):
        import torch
        """Tests _execute_mm_encoder with multiple requests and images."""
        # 1. ===== Setup =====
        self.runner.is_multimodal_model = True
        self.mock_get_mm_embed_fn = MagicMock()
        self.runner.embed_multimodal_fn = self.mock_get_mm_embed_fn

        self.runner.state = MagicMock()
        # Mock scheduler output for two requests
        mock_scheduler_output = MagicMock(spec=VllmSchedulerOutput)
        mock_scheduler_output.scheduled_encoder_inputs = {
            "req-1": [0],
            "req-2": [0]
        }

        # Mock request states
        px_1 = torch.randn(3, 224, 224, dtype=torch.bfloat16)
        grid_1 = torch.tensor([[1, 1, 1]], dtype=torch.int64)

        mm_item_1 = MultiModalKwargsItem({
            "pixel_values":
            MultiModalFieldElem(px_1, MultiModalBatchedField()),
            "image_grid_thw":
            MultiModalFieldElem(grid_1, MultiModalBatchedField())
        })

        req_state_1 = CachedRequestState(
            req_id="req-1",
            prompt_token_ids=[],
            output_token_ids=[],
            sampling_params=MagicMock(),
            block_ids=(),
            num_computed_tokens=0,
            mm_features=[
                MultiModalFeatureSpec(data=mm_item_1,
                                      identifier="req-1",
                                      modality="image",
                                      mm_position=PlaceholderRange(offset=0,
                                                                   length=1))
            ],
            lora_request=None,
            pooling_params=None,
            generator=None)

        px_2 = torch.randn(3, 224, 224, dtype=torch.bfloat16)
        grid_2 = torch.tensor([[1, 2, 2]], dtype=torch.int64)
        mm_item_2 = MultiModalKwargsItem({
            "pixel_values":
            MultiModalFieldElem(px_2, MultiModalBatchedField()),
            "image_grid_thw":
            MultiModalFieldElem(grid_2, MultiModalBatchedField())
        })

        req_state_2 = CachedRequestState(
            req_id="req-2",
            prompt_token_ids=[],
            output_token_ids=[],
            sampling_params=MagicMock(),
            block_ids=(),
            num_computed_tokens=0,
            mm_features=[
                MultiModalFeatureSpec(data=mm_item_2,
                                      identifier="req-2",
                                      modality="image",
                                      mm_position=PlaceholderRange(offset=0,
                                                                   length=1))
            ],
            lora_request=None,
            pooling_params=None,
            generator=None)

        self.runner.requests = {"req-1": req_state_1, "req-2": req_state_2}

        emb_1 = jnp.ones((10, 128), dtype=jnp.bfloat16)
        emb_2 = jnp.ones((20, 128), dtype=jnp.bfloat16) * 2
        self.mock_get_mm_embed_fn.return_value = (emb_1, emb_2)

        # 2. ===== Act =====
        self.runner.mm_manager.execute_mm_encoder(mock_scheduler_output)

        # 3. ===== Assert =====
        assert "req-1" in self.runner.encoder_cache
        np.testing.assert_array_equal(
            np.asarray(self.runner.encoder_cache["req-1"]), np.asarray(emb_1))
        assert "req-2" in self.runner.encoder_cache
        np.testing.assert_array_equal(
            np.asarray(self.runner.encoder_cache["req-2"]), np.asarray(emb_2))

        self.mock_get_mm_embed_fn.assert_called_once()
        call_args = self.mock_get_mm_embed_fn.call_args

        state_arg, grid_arg = call_args.args
        kwargs_arg = call_args.kwargs

        assert state_arg == self.runner.state
        assert grid_arg == ((1, 1, 1), (1, 2, 2))
        assert "pixel_values" in kwargs_arg

        passed_pixel_values = kwargs_arg['pixel_values']
        assert passed_pixel_values.shape == (2, 3, 224, 224)

        expected_pixel_values = torch.stack([px_1, px_2], dim=0).to(
            torch.float32).numpy().astype(jnp.bfloat16)
        np.testing.assert_array_equal(np.asarray(passed_pixel_values),
                                      expected_pixel_values)

    def test_gather_mm_embeddings_chunked_prefill(self):
        """Tests _gather_mm_embeddings with chunked prefill scenarios."""
        # 1. ===== Setup =====
        self.runner.is_multimodal_model = True
        req_id = "req-1"

        # Mock encoder output
        encoder_embedding = jnp.arange(56 * 128, dtype=jnp.bfloat16).reshape(
            (56, 128))
        self.runner.encoder_cache = {req_id: encoder_embedding}

        mock_sampling_params = MagicMock()
        mock_sampling_params.sampling_type = SamplingType.GREEDY
        mock_sampling_params.top_k = -1
        mock_sampling_params.top_p = 1.0
        mock_sampling_params.temperature = 0.0
        mock_sampling_params.min_tokens = 0
        mock_sampling_params.logprobs = None
        mock_sampling_params.logit_bias = None
        mock_sampling_params.allowed_token_ids = set()
        mock_sampling_params.bad_words_token_ids = None
        mock_sampling_params.all_stop_token_ids = set()

        # Mock request state
        prompt_token_len = 100
        mm_position = PlaceholderRange(offset=10, length=56)
        is_mm_embed_cpu = np.zeros(prompt_token_len, dtype=np.bool_)
        is_mm_embed_cpu[mm_position.offset:mm_position.offset +
                        mm_position.length] = True
        req_state = CachedRequestState(
            req_id=req_id,
            prompt_token_ids=list(range(prompt_token_len)),
            output_token_ids=[],
            sampling_params=mock_sampling_params,
            block_ids=([], ),
            num_computed_tokens=0,  # This will be updated per step
            mm_features=[
                MultiModalFeatureSpec(data=None,
                                      identifier=req_id,
                                      modality="image",
                                      mm_position=mm_position)
            ],
            lora_request=None,
            pooling_params=None,
            generator=None,
        )
        self.runner.requests = {req_id: req_state}
        self.runner.input_batch.add_request(req_state)

        # 2. ===== Act & Assert =====

        # ----- Step 1: First chunk of prefill -----
        req_state.num_computed_tokens = 0
        mock_scheduler_output_1 = MagicMock(spec=VllmSchedulerOutput)
        mock_scheduler_output_1.total_num_scheduled_tokens = 20
        mock_scheduler_output_1.num_scheduled_tokens = {req_id: 20}

        gathered_embeds_1_padded, gathered_is_mm_embed_1 = self.runner.mm_manager.gather_mm_embeddings(
            mock_scheduler_output_1,
            target_pad_len=mock_scheduler_output_1.total_num_scheduled_tokens)

        assert gathered_embeds_1_padded is not None
        assert gathered_is_mm_embed_1 is not None
        expected_embeds_1 = encoder_embedding[0:10]
        assert gathered_embeds_1_padded.shape[
            0] == mock_scheduler_output_1.total_num_scheduled_tokens
        gathered_embeds_1 = gathered_embeds_1_padded[:10]
        assert gathered_embeds_1.shape == expected_embeds_1.shape
        np.testing.assert_array_equal(np.asarray(gathered_embeds_1),
                                      np.asarray(expected_embeds_1))
        assert gathered_is_mm_embed_1.shape == (20, )
        np.testing.assert_array_equal(np.asarray(gathered_is_mm_embed_1),
                                      is_mm_embed_cpu[:20])

        # ----- Step 2: Middle chunk of prefill -----
        req_state.num_computed_tokens = 20
        mock_scheduler_output_2 = MagicMock(spec=VllmSchedulerOutput)
        mock_scheduler_output_2.total_num_scheduled_tokens = 30
        mock_scheduler_output_2.num_scheduled_tokens = {req_id: 30}

        gathered_embeds_2_padded, gathered_is_mm_embed_2 = self.runner.mm_manager.gather_mm_embeddings(
            mock_scheduler_output_2,
            target_pad_len=mock_scheduler_output_2.total_num_scheduled_tokens)

        assert gathered_embeds_2_padded is not None
        assert gathered_is_mm_embed_2 is not None
        expected_embeds_2 = encoder_embedding[10:40]
        assert gathered_embeds_2_padded.shape[
            0] == mock_scheduler_output_2.total_num_scheduled_tokens
        gathered_embeds_2 = gathered_embeds_2_padded[:30]
        np.testing.assert_array_equal(np.asarray(gathered_embeds_2),
                                      np.asarray(expected_embeds_2))
        assert gathered_is_mm_embed_2.shape == (30, )
        np.testing.assert_array_equal(np.asarray(gathered_is_mm_embed_2),
                                      is_mm_embed_cpu[20:50])

        # ----- Step 3: Last chunk of prefill -----
        req_state.num_computed_tokens = 50
        mock_scheduler_output_3 = MagicMock(spec=VllmSchedulerOutput)
        mock_scheduler_output_3.total_num_scheduled_tokens = 30
        mock_scheduler_output_3.num_scheduled_tokens = {req_id: 30}

        gathered_embeds_3_padded, gathered_is_mm_embed_3 = self.runner.mm_manager.gather_mm_embeddings(
            mock_scheduler_output_3,
            target_pad_len=mock_scheduler_output_3.total_num_scheduled_tokens)

        assert gathered_embeds_3_padded is not None
        assert gathered_is_mm_embed_3 is not None
        expected_embeds_3 = encoder_embedding[40:56]
        assert gathered_embeds_3_padded.shape[
            0] == mock_scheduler_output_3.total_num_scheduled_tokens
        gathered_embeds_3 = gathered_embeds_3_padded[:16]
        np.testing.assert_array_equal(np.asarray(gathered_embeds_3),
                                      np.asarray(expected_embeds_3))
        assert gathered_is_mm_embed_3.shape == (30, )
        np.testing.assert_array_equal(np.asarray(gathered_is_mm_embed_3),
                                      is_mm_embed_cpu[50:80])

    def test_calc_mrope_positions(self):
        """Tests the calculation of M-RoPE positions for mixed prompt/completion."""
        # 1. ===== Setup =====
        self.runner.uses_mrope = True
        req_id = "req-1"
        prompt_len = 20
        num_computed = 15
        num_scheduled = 10
        mrope_delta = 100

        # Mock request state with pre-computed mrope positions for the prompt
        mock_mrope_positions = np.arange(3 * prompt_len,
                                         dtype=np.int64).reshape(
                                             3, prompt_len)
        mock_sampling_params = MagicMock()
        mock_sampling_params.sampling_type = SamplingType.GREEDY
        mock_sampling_params.top_k = -1
        mock_sampling_params.top_p = 1.0
        mock_sampling_params.temperature = 0.0
        mock_sampling_params.min_tokens = 0
        mock_sampling_params.logprobs = None
        mock_sampling_params.logit_bias = None
        mock_sampling_params.allowed_token_ids = set()
        mock_sampling_params.bad_words_token_ids = None
        mock_sampling_params.all_stop_token_ids = set()

        req_state = CachedRequestState(
            req_id=req_id,
            prompt_token_ids=list(range(prompt_len)),
            output_token_ids=[],
            sampling_params=mock_sampling_params,
            block_ids=([], ),
            num_computed_tokens=num_computed,
            mm_features=[],
            lora_request=None,
            pooling_params=None,
            generator=None,
            mrope_positions=mock_mrope_positions,
            mrope_position_delta=mrope_delta,
        )
        self.runner.requests = {req_id: req_state}
        self.runner.input_batch.add_request(req_state)
        # Manually set num_computed_tokens in the batch as add_request sets it to 0
        self.runner.input_batch.num_computed_tokens_cpu[0] = num_computed

        # Mock scheduler output
        mock_scheduler_output = MagicMock(spec=VllmSchedulerOutput)
        mock_scheduler_output.num_scheduled_tokens = {req_id: num_scheduled}

        # Patch the static method that computes completion positions
        with patch.object(MRotaryEmbedding,
                          "get_next_input_positions_tensor") as mock_get_next:
            # 2. ===== Act =====
            self.runner.mm_manager.calc_mrope_positions(mock_scheduler_output)

            # 3. ===== Assert =====
            # The first 5 positions should be copied from the pre-computed prompt positions
            expected_prompt_part = mock_mrope_positions[:, 15:20]
            actual_prompt_part = self.runner.mrope_positions_cpu[:, 0:5]
            np.testing.assert_array_equal(actual_prompt_part,
                                          expected_prompt_part)

            # The next 5 positions should be computed on-the-fly
            mock_get_next.assert_called_once()
            call_kwargs = mock_get_next.call_args.kwargs
            np.testing.assert_array_equal(call_kwargs["out"],
                                          self.runner.mrope_positions_cpu)
            assert call_kwargs["out_offset"] == 5
            assert call_kwargs["mrope_position_delta"] == mrope_delta
            assert call_kwargs["context_len"] == prompt_len
            assert call_kwargs["num_new_tokens"] == 5



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from typing import Optional

import jax
import jax.numpy as jnp
import torch
from compressed_tensors.quantization import QuantizationStrategy
from jax.sharding import PartitionSpec
from torch.nn.parameter import Parameter
from torchax.interop import jax_view, torch_view
from vllm.model_executor.layers.quantization.compressed_tensors.schemes.compressed_tensors_w8a8_int8 import \
    CompressedTensorsW8A8Int8
from vllm.model_executor.layers.quantization.utils.w8a8_utils import \
    convert_to_channelwise
from vllm.model_executor.parameter import (BasevLLMParameter,
                                           ChannelQuantScaleParameter,
                                           ModelWeightParameter,
                                           PerTensorScaleParameter)

from tpu_inference.layers.common.linear import sharded_quantized_matmul
from tpu_inference.layers.common.process_weights.linear_weights import (
    LinearWeights, process_linear_weights, shard_linear_weights,
    to_parameter_list)
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation
from tpu_inference.layers.vllm.quantization.configs import \
    VllmQuantLinearConfig
from tpu_inference.logger import init_logger
from tpu_inference.utils import t2j

P = PartitionSpec
logger = init_logger(__name__)


class VllmCompressedTensorsW8A8Int8(CompressedTensorsW8A8Int8):

    def __init__(self, strategy: str, is_static_input_scheme: bool,
                 input_symmetric: bool, linear_config: VllmQuantLinearConfig):
        super().__init__(strategy, is_static_input_scheme, input_symmetric)

        self.linear_config = linear_config
        self.is_channelwise = (self.strategy == QuantizationStrategy.CHANNEL)

    def create_weights(
        self,
        layer: torch.nn.Module,
        output_partition_sizes: list[int],
        input_size_per_partition: int,
        params_dtype: torch.dtype,
        weight_loader: Callable,
        **kwargs,
    ):
        layer.logical_widths = output_partition_sizes

        # WEIGHT
        weight = ModelWeightParameter(
            data=torch.empty(sum(output_partition_sizes),
                             input_size_per_partition,
                             dtype=torch.int8),
            input_dim=1,
            output_dim=0,
            weight_loader=weight_loader,
        )

        layer.register_parameter("weight", weight)

        # WEIGHT SCALE
        if self.strategy == QuantizationStrategy.CHANNEL:
            weight_scale = ChannelQuantScaleParameter(
                data=torch.empty((sum(output_partition_sizes), 1),
                                 dtype=torch.float32),
                output_dim=0,
                weight_loader=weight_loader,
            )
        else:
            assert self.strategy == QuantizationStrategy.TENSOR
            weight_scale = PerTensorScaleParameter(
                data=torch.empty(len(output_partition_sizes),
                                 dtype=torch.float32),
                weight_loader=weight_loader,
            )
        layer.register_parameter("weight_scale", weight_scale)

        # INPUT SCALE
        if self.is_static_input_scheme:
            input_scale = BasevLLMParameter(data=torch.empty(
                1, dtype=torch.float32),
                                            weight_loader=weight_loader)
            layer.register_parameter("input_scale", input_scale)

            if not self.input_symmetric:
                # Note: compressed-tensors stores the zp using the same dtype
                # as the weights
                # AZP loaded as int8 but used as int32
                input_zero_point = BasevLLMParameter(
                    data=torch.empty(1, dtype=torch.int8),
                    weight_loader=weight_loader)
                layer.register_parameter("input_zero_point", input_zero_point)

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        weight = t2j(layer.weight, use_dlpack=False)
        delattr(layer, "weight")

        weight_scale = layer.weight_scale
        is_fused_module = len(layer.logical_widths) > 1
        if is_fused_module and not self.is_channelwise:
            weight_scale = convert_to_channelwise(weight_scale,
                                                  layer.logical_widths)
        weight_scale = weight_scale.squeeze(-1)
        weight_scale = t2j(weight_scale, use_dlpack=False)
        delattr(layer, "weight_scale")

        if layer.bias is not None and not layer.skip_bias_add:
            if layer.return_bias:
                logger.warning_once("Bias might return incorrect value.")
            bias = t2j(layer.bias, use_dlpack=False)
            delattr(layer, "bias")
        else:
            bias = None

        @jax.jit
        def process_int8_linear_weights(
            weight: jax.Array,
            weight_scale: jax.Array,
            bias: jax.Array | None,
        ) -> LinearWeights:
            return process_linear_weights(
                LinearWeights(
                    weight=weight,
                    weight_scale=weight_scale,
                    zero_point=None,
                    bias=bias,
                ),
                fused=self.linear_config.fuse_matmuls,
                output_sizes=self.linear_config.output_sizes,
                reorder_size=self.linear_config.n_shards,
            )

        weights = process_int8_linear_weights(weight, weight_scale, bias)
        weights = torch_view(
            shard_linear_weights(
                weights,
                mesh=self.linear_config.mesh,
                weight_p_spec=self.linear_config.weight_sharding,
                bias_p_spec=self.linear_config.bias_sharding,
            ))

        if self.linear_config.fuse_matmuls:
            layer.weight = Parameter(weights.weight, requires_grad=False)
            layer.weight_scale = Parameter(weights.weight_scale,
                                           requires_grad=False)
            if bias is not None:
                layer.bias = Parameter(weights.bias, requires_grad=False)
        else:
            layer.weight = to_parameter_list(weights.weight)
            layer.weight_scale = to_parameter_list(weights.weight_scale)
            if bias is not None:
                layer.bias = to_parameter_list(weights.bias)

        # TODO(kyuyeunk): Support static range input quantization.
        assert getattr(layer, "input_scale", None) is None
        assert getattr(layer, "input_zero_point", None) is None
        assert getattr(layer, "azp_adj", None) is None

    def apply_weights(self, layer: torch.nn.Module, x: torch.Tensor,
                      bias: Optional[torch.Tensor]) -> torch.Tensor:
        with jax.named_scope(layer._get_name()):
            if self.linear_config.fuse_matmuls:
                out = self._apply_fused(layer, x, bias)
            else:
                out = self._apply_split(layer, x, bias)

        return out

    def _apply_fused(self, layer: torch.nn.Module, x: torch.Tensor,
                     bias: Optional[torch.Tensor]) -> torch.Tensor:
        x_jax = jax_view(x)
        weight_jax = jax_view(layer.weight)
        weight_scale_jax = jax_view(layer.weight_scale)

        outs = sharded_quantized_matmul(
            x_jax,
            weight_jax,
            weight_scale_jax,
            self.linear_config.weight_sharding,
            mesh=self.linear_config.mesh,
        )
        if bias is not None and not layer.skip_bias_add:
            outs += jax_view(bias)

        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        out = jnp.concatenate(outs, axis=-1)
        return torch_view(out)

    def _apply_split(self, layer: torch.nn.Module, x: torch.Tensor,
                     bias: Optional[torch.Tensor]) -> torch.Tensor:
        assert isinstance(layer.weight, torch.nn.ParameterList)

        x_jax = jax_view(x)
        outs = []
        for i, (weight, weight_scale) in enumerate(
                zip(layer.weight, layer.weight_scale)):
            weight_jax = jax_view(weight)
            weight_scale_jax = jax_view(weight_scale)

            out = sharded_quantized_matmul(
                x_jax,
                weight_jax,
                weight_scale_jax,
                self.linear_config.weight_sharding,
                mesh=self.linear_config.mesh,
            )
            if bias is not None and not layer.skip_bias_add:
                out += jax_view(bias[i])

            outs.append(out)
        out = jnp.concatenate(outs, axis=-1)
        return torch_view(out)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import tempfile

import jax
import pytest
import torch
import torch.nn.functional as F
import torchax
from jax._src import test_util as jtu
from jax.sharding import PartitionSpec
from torchax.interop import jax_view, torch_view
from torchax.ops.mappings import j2t, t2j
from vllm.config import ParallelConfig, set_current_vllm_config
from vllm.distributed.parallel_state import (ensure_model_parallel_initialized,
                                             init_distributed_environment)
from vllm.engine.arg_utils import EngineArgs
from vllm.forward_context import set_forward_context
from vllm.model_executor.layers.fused_moe import FusedMoE
from vllm.model_executor.layers.linear import (ColumnParallelLinear,
                                               LinearBase,
                                               MergedColumnParallelLinear,
                                               QKVParallelLinear,
                                               RowParallelLinear)

from tests.layers.common import utils as test_utils
from tpu_inference.layers.common.moe import MoEBackend
from tpu_inference.layers.common.quantization.configs import QuantLinearConfig
from tpu_inference.layers.vllm.quantization import get_tpu_quantization_config
from tpu_inference.layers.vllm.quantization.configs import \
    VllmQuantLinearConfig
from tpu_inference.layers.vllm.quantization.fp8 import (VllmFp8Config,
                                                        VllmFp8LinearMethod,
                                                        VllmFp8MoEMethod)

P = PartitionSpec
MODELS = [
    "Qwen/Qwen3-0.6B-FP8",
]


def ref_quantize_fp8(x: torch.Tensor,
                     dtype: torch.dtype,
                     per_tensor: bool = False):
    dtype_info = torch.finfo(dtype)
    dtype_max = float(dtype_info.max)
    dtype_min = float(dtype_info.min)

    dim = () if per_tensor else 1
    x_abs_max = torch.amax(torch.abs(x), dim=dim, keepdim=True)
    if per_tensor:
        x_abs_max = torch.squeeze(x_abs_max, dim=-1)
    x_s = x_abs_max / dtype_max
    x_q = torch.clip(x / x_s, dtype_min, dtype_max).to(dtype)
    return x_q, x_s.to(torch.float32)


def ref_fp8_activation(
        x: torch.Tensor,
        dtype: torch.dtype = torch.float8_e4m3fn) -> torch.Tensor:
    x_q, x_scale = ref_quantize_fp8(x, dtype=dtype, per_tensor=False)
    return x_q.to(torch.float32) * x_scale


def ref_dequantize_fp8_block_2d(w_q: torch.Tensor, scale_blocks: torch.Tensor,
                                block_m: int, block_n: int) -> torch.Tensor:
    out, inn = w_q.shape
    scale_out, scale_inn = scale_blocks.shape
    padded_out, padded_inn = scale_out * block_m, scale_inn * block_n

    w_q = F.pad(w_q, (0, padded_inn - inn, 0, padded_out - out))

    w_q = w_q.to(torch.float32).view(scale_out, block_m, scale_inn, block_n)
    scale_e = scale_blocks[:, None, :, None]
    w_deq = w_q * scale_e

    w_deq = w_deq.reshape(padded_out, padded_inn)
    return w_deq[:out, :inn]


def return_ref_and_layer_output(layer: torch.nn.Module, batch_size: int = 16):
    assert isinstance(layer, LinearBase)
    quant_method = layer.quant_method
    assert isinstance(quant_method, VllmFp8LinearMethod)
    quant_method.requant_block_size = None
    quant_method.requant_weight_dtype = jax.numpy.float8_e4m3fn

    input_tensor = torch.rand(
        batch_size, layer.input_size, dtype=torch.bfloat16) / 10
    input_tensor = input_tensor.to('cpu')

    x_deq = ref_fp8_activation(input_tensor, torch.float8_e4m3fn)

    assert hasattr(layer.quant_method, "weight_block_size")
    wbs = layer.quant_method.weight_block_size
    block_m, block_n = int(wbs[0]), int(wbs[1])

    w_deq = ref_dequantize_fp8_block_2d(layer.weight.data,
                                        layer.weight_scale_inv.data, block_m,
                                        block_n)

    ref_output = torch.einsum('bd,fd->bf', x_deq.to(torch.float32),
                              w_deq.to(torch.float32))

    if layer.bias is not None:
        ref_output = ref_output + layer.bias.data

    ref_output = ref_output.to(input_tensor.dtype)

    with torchax.default_env():
        quant_method.process_weights_after_loading(layer)

        jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
        layer_output = layer(jax_input_tensor)
        layer_output = j2t(layer_output.to(torch.float32)).to(torch.bfloat16)

    return ref_output, layer_output


def ref_quantize_fp8_block_2d(w: torch.Tensor, block_m: int, block_n: int,
                              dtype: torch.dtype):
    dtype_info = torch.finfo(dtype)
    dtype_max = float(dtype_info.max)
    dtype_min = float(dtype_info.min)

    out, inn = w.shape
    scale_out, scale_inn = math.ceil(out / block_m), math.ceil(inn / block_n)
    padded_out, padded_inn = scale_out * block_m, scale_inn * block_n

    w = F.pad(w, (0, padded_inn - inn, 0, padded_out - out))
    w_view = w.view(scale_out, block_m, scale_inn, block_n)

    abs_max = torch.amax(torch.abs(w_view), dim=(1, 3), keepdim=True)
    scale = abs_max / dtype_max
    w_q = torch.clamp(w_view / scale, dtype_min, dtype_max).to(dtype)

    w_q = w_q.reshape(padded_out, padded_inn)
    w_q = w_q[:out, :inn]

    scale_blocks = scale.squeeze(1).squeeze(-1).to(torch.float32)
    return w_q, scale_blocks


def quantize_to_fp8_block_3d(weight: torch.Tensor, block_m: int, block_n: int,
                             dtype: torch.dtype):
    dtype_info = torch.finfo(dtype)
    dtype_max = float(dtype_info.max)
    dtype_min = float(dtype_info.min)

    num_experts, out, inn = weight.shape

    assert out % block_m == 0 and inn % block_n == 0

    weight_view = weight.view(num_experts, out // block_m, block_m,
                              inn // block_n, block_n)

    abs_max = torch.amax(torch.abs(weight_view), dim=(2, 4), keepdim=True)
    scale = abs_max / dtype_max
    w_q = torch.clamp(weight_view / scale, dtype_min, dtype_max).to(dtype)

    w_q = w_q.reshape(num_experts, out, inn)
    scale_blocks = scale.squeeze(2).squeeze(-1).to(torch.float32)
    return w_q, scale_blocks


def initialize_layer_weights(layer: torch.nn.Module):
    assert isinstance(layer, LinearBase)
    assert isinstance(layer.quant_method, VllmFp8LinearMethod)
    layer.quant_method.linear_config.requant_block_size = None
    layer.quant_method.linear_config.requant_weight_dtype = jax.numpy.float8_e4m3fn

    assert hasattr(layer.quant_method, "weight_block_size")
    block_m, block_n = layer.quant_method.weight_block_size

    w_f32 = (
        torch.rand(layer.output_size, layer.input_size, dtype=torch.float32) /
        10)

    w_q, w_scale_blocks = ref_quantize_fp8_block_2d(w_f32, block_m, block_n,
                                                    torch.float8_e4m3fn)

    layer.weight.data = w_q
    assert hasattr(layer, "weight_scale_inv")

    layer.weight_scale_inv.data = w_scale_blocks
    assert layer.weight_scale_inv.data.shape == w_scale_blocks.shape

    if layer.bias is not None:
        layer.bias.data = torch.rand_like(layer.bias.data) / 10.0


@pytest.fixture(autouse=True)
def setup_environment():
    # This is a fake config used for init dist env.
    # RowParallelLinear needs dist env to be initialized.
    engine_args = EngineArgs(model=MODELS[0],
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)

    vllm_config = engine_args.create_engine_config()

    with set_current_vllm_config(vllm_config):
        temp_file = tempfile.mkstemp()[1]
        init_distributed_environment(
            1,
            0,
            local_rank=0,
            distributed_init_method=f"file://{temp_file}",
            backend="gloo")
        ensure_model_parallel_initialized(1, 1)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
def test_quant_override(model, mesh):

    engine_args = EngineArgs(model=model,
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = torch.bfloat16

    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    assert isinstance(quant_config, VllmFp8Config)
    assert quant_config.vllm_config == vllm_config
    assert quant_config.mesh == mesh


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_row_parallel_linear(model, bias, num_devices, enable_sp,
                             enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(model=model,
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    vllm_config.model_config.dtype = dtype
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = RowParallelLinear(
            input_size=4096,
            output_size=8192,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    initialize_layer_weights(linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(linear_layer)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_column_parallel_linear(model, bias, num_devices, enable_sp,
                                enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(model=model,
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    vllm_config.model_config.dtype = dtype
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = ColumnParallelLinear(
            input_size=4096,
            output_size=8192,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    initialize_layer_weights(linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(linear_layer)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("fuse_matmuls", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_qkv_parallel_linear(model, bias, num_devices, enable_sp, fuse_matmuls,
                             enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(model=model,
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = QKVParallelLinear(
            hidden_size=4096,
            head_size=128,
            total_num_heads=32,
            total_num_kv_heads=8,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )
        assert isinstance(linear_layer.quant_method.linear_config,
                          QuantLinearConfig)
        linear_layer.quant_method.linear_config.fuse_matmuls = fuse_matmuls

    initialize_layer_weights(linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(linear_layer)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("fuse_matmuls", [False, True])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_merged_column_parallel_linear(model, bias, num_devices, fuse_matmuls,
                                       enable_sp, enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(model=model,
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = MergedColumnParallelLinear(
            input_size=4096,
            output_sizes=[14336] * 2,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )
        assert isinstance(linear_layer.quant_method.linear_config,
                          QuantLinearConfig)
        linear_layer.quant_method.linear_config.fuse_matmuls = fuse_matmuls

    initialize_layer_weights(linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(linear_layer)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("use_ep", [True, False])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("num_tokens", [8, 32])
@pytest.mark.parametrize("intermediate_size", [1024, 2048])
@pytest.mark.parametrize("hidden_size", [128, 512])
@pytest.mark.parametrize("num_experts", [8])
@pytest.mark.parametrize("topk", [2])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_fused_moe(use_ep, num_devices, num_tokens, intermediate_size,
                   hidden_size, num_experts, topk, enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)

    torch.manual_seed(42)
    dtype = torch.bfloat16

    a = torch.randn((num_tokens, hidden_size), dtype=dtype) / 10
    w1 = torch.randn(
        (num_experts, 2 * intermediate_size, hidden_size), dtype=dtype) / 10
    w2 = torch.randn(
        (num_experts, hidden_size, intermediate_size), dtype=dtype) / 10
    score = torch.randn((num_tokens, num_experts), dtype=dtype)

    engine_args = EngineArgs(model=MODELS[0],
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = dtype
    vllm_config.parallel_config = ParallelConfig(
        tensor_parallel_size=mesh.devices.size, enable_expert_parallel=use_ep)

    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        vllm_fused_moe = FusedMoE(
            num_experts=num_experts,
            top_k=topk,
            hidden_size=hidden_size,
            intermediate_size=intermediate_size,
            reduce_results=False,
            renormalize=False,
            tp_size=1,
            dp_size=1,
            quant_config=quant_config,
            has_bias=False,
        )
        vllm_fused_moe.moe_parallel_config.use_ep = use_ep

    block_m, block_n = vllm_fused_moe.quant_method.quant_config.weight_block_size

    w1_weight, w1_weight_scale = quantize_to_fp8_block_3d(
        w1, block_m, block_n, torch.float8_e4m3fn)
    w2_weight, w2_weight_scale = quantize_to_fp8_block_3d(
        w2, block_m, block_n, torch.float8_e4m3fn)

    vllm_fused_moe.w13_weight.data = w1_weight
    vllm_fused_moe.w2_weight.data = w2_weight
    vllm_fused_moe.w13_weight_scale_inv.data = w1_weight_scale
    vllm_fused_moe.w2_weight_scale_inv.data = w2_weight_scale

    expected = test_utils.ref_moe(a, score, w1, w2, None, None,
                                  vllm_fused_moe.top_k,
                                  vllm_fused_moe.renormalize,
                                  vllm_fused_moe.activation.value)

    with torchax.default_env(), set_forward_context(None, vllm_config):
        assert isinstance(vllm_fused_moe.quant_method, VllmFp8MoEMethod)
        if use_ep:
            assert vllm_fused_moe.quant_method.moe_backend == MoEBackend.GMM_EP
        else:
            assert vllm_fused_moe.quant_method.moe_backend == MoEBackend.GMM_TP

        jax_a = a.to('jax')
        jax_score = score.to('jax')

        vllm_fused_moe.quant_method.process_weights_after_loading(
            vllm_fused_moe)
        actual = vllm_fused_moe(jax_a, jax_score)

        torch.testing.assert_close(expected,
                                   actual,
                                   check_device=False,
                                   atol=2e-2,
                                   rtol=0.0)


@pytest.mark.parametrize("requant_block_size", (128, 512))
@pytest.mark.parametrize("requant_weight_dtype",
                         (jax.numpy.float8_e4m3fn, jax.numpy.float4_e2m1fn))
def test_blockwise_quant(requant_block_size, requant_weight_dtype):
    if not jtu.is_device_tpu_at_least(version=7):
        pytest.skip("Expect TPUv7+")
    mesh = test_utils.get_spmd_mesh()
    dtype = torch.bfloat16

    engine_args = EngineArgs(model=MODELS[0],
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()

    vllm_config.model_config.dtype = dtype
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = RowParallelLinear(
            input_size=4096,
            output_size=5120,
            bias=False,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    initialize_layer_weights(linear_layer)
    quant_method = linear_layer.quant_method
    quant_method.linear_config.requant_block_size = requant_block_size
    quant_method.linear_config.requant_weight_dtype = requant_weight_dtype
    quant_method.linear_config.enable_quantized_matmul_kernel = True

    quant_method.process_weights_after_loading(linear_layer)
    weight_jax = jax_view(linear_layer.weight)
    weight_scale_jax = jax_view(linear_layer.weight_scale)
    assert weight_jax.shape == (5120, 4096)
    assert weight_scale_jax.shape == (4096 // requant_block_size, 1, 5120)
    assert weight_jax.dtype == requant_weight_dtype

    # TODO: Check output similarity between quantized and unquantized ones.
    input_tensor = torch.rand(
        16, linear_layer.input_size, dtype=torch.bfloat16) / 10
    input_tensor = input_tensor.to('cpu')
    jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
    quantized_output = linear_layer(jax_input_tensor)
    assert quantized_output.shape == (16, 5120)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("input_size,output_size", [
    (7168, 576),
])
def test_unaligned_block_quantization(model, input_size, output_size):
    mesh = test_utils.get_spmd_mesh(1)

    torch.manual_seed(42)
    dtype = torch.bfloat16

    engine_args = EngineArgs(model=model,
                             max_model_len=64,
                             max_num_batched_tokens=64,
                             max_num_seqs=4,
                             trust_remote_code=True)
    vllm_config = engine_args.create_engine_config()

    vllm_config.model_config.dtype = dtype
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = RowParallelLinear(
            input_size=input_size,
            output_size=output_size,
            bias=False,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    initialize_layer_weights(linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(linear_layer)
    torch.testing.assert_close(ref_output, layer_output)


def test_fp8_init():
    """Verify that vLLM patches are working for __init__ post vLLM PR: https://github.com/vllm-project/vllm/pull/32929."""
    mesh = test_utils.get_spmd_mesh(1)
    engine_args = EngineArgs(model=MODELS[0])
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)

    # VllmQuantLinearConfig requires a layer to initialize.
    with set_current_vllm_config(vllm_config):
        dummy_layer = RowParallelLinear(
            input_size=128,
            output_size=128,
            bias=False,
            params_dtype=torch.bfloat16,
            quant_config=quant_config,
        )
    linear_config = VllmQuantLinearConfig(vllm_config, mesh, dummy_layer)

    with set_current_vllm_config(vllm_config):
        method = VllmFp8LinearMethod(quant_config, linear_config)

    assert method.fp8_linear is None, "fp8_linear is now set to None"
    assert method.use_marlin is True, "use_marlin is set to True (previous status quo)"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import jax
import jax.numpy as jnp
import torch
from jax.sharding import Mesh, PartitionSpec
from torch.nn.parameter import Parameter
from torchax.interop import jax_view, torch_view
from vllm.model_executor.layers import linear as vllm_linear
from vllm.model_executor.layers.attention import Attention
from vllm.model_executor.layers.fused_moe import FusedMoE, FusedMoEMethodBase
from vllm.model_executor.layers.quantization import fp8 as vllm_fp8
from vllm.model_executor.layers.quantization import \
    register_quantization_config
from vllm.model_executor.layers.quantization.base_config import \
    QuantizeMethodBase
from vllm.model_executor.layers.quantization.utils.quant_utils import \
    is_layer_skipped

from tpu_inference.layers.common.moe import MoEBackend
from tpu_inference.layers.common.process_weights.linear_weights import (
    shard_linear_weights, to_parameter_list)
from tpu_inference.layers.common.process_weights.moe_weights import (
    FusedMoEWeights, process_fp8_moe_weights, shard_moe_weights)
from tpu_inference.layers.common.quant_methods import FP8
from tpu_inference.layers.common.quantization import fp8 as common_fp8
from tpu_inference.layers.vllm.interface.moe import (
    select_moe_backend_from_fused_moe_config, vllm_moe_apply)
from tpu_inference.layers.vllm.quantization.configs import (
    VllmQuantConfig, VllmQuantLinearConfig)
from tpu_inference.layers.vllm.quantization.unquantized import (
    VllmUnquantizedFusedMoEMethod, VllmUnquantizedLinearMethod)
from tpu_inference.logger import init_logger
from tpu_inference.utils import t2j

P = PartitionSpec

logger = init_logger(__name__)


@register_quantization_config(FP8)
class VllmFp8Config(vllm_fp8.Fp8Config, VllmQuantConfig):

    @classmethod
    def get_name(cls):
        return FP8

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional[Union[vllm_linear.LinearMethodBase, QuantizeMethodBase]]:
        match layer:
            case vllm_linear.LinearBase():
                linear_config = self.get_linear_config(layer)
                if is_layer_skipped(
                        prefix=prefix,
                        ignored_layers=self.ignored_layers,
                        fused_mapping=self.packed_modules_mapping,
                ):
                    return VllmUnquantizedLinearMethod(linear_config)
                return VllmFp8LinearMethod(self, linear_config)
            case FusedMoE():
                if is_layer_skipped(
                        prefix=prefix,
                        ignored_layers=self.ignored_layers,
                        fused_mapping=self.packed_modules_mapping,
                ):
                    return VllmUnquantizedFusedMoEMethod(layer.moe_config)
                if self.is_checkpoint_fp8_serialized:
                    layer.moe_config = self.get_moe_config(layer)
                    return VllmFp8MoEMethod(self, layer, self.mesh)
                else:
                    raise NotImplementedError(
                        "FP8OnelineMoEMethod is not supported.")
            case Attention():
                logger.warning_once("FP8KVCacheMethod is not implemented. "
                                    "Skipping quantization for this layer.")
                return None
            case _:
                return None


class VllmFp8LinearMethod(vllm_fp8.Fp8LinearMethod,
                          common_fp8.Fp8LinearMethod):

    def __init__(
        self,
        quant_config: VllmFp8Config,
        linear_config: VllmQuantLinearConfig,
    ):

        # Per https://github.com/vllm-project/vllm/pull/32929,
        # init_fp8_linear_kernel is now called by super().__init__
        # but does not support TPU backends as expected.
        # use_marlin was also changed to be determined via isinstance(self.fp8_linear, MarlinFP8ScaledMMLinearKernel).
        # We need to monkeypatch init_fp8_linear_kernel and explicitly set use_marlin = True
        # in order to bypass using native vLLM's vllm/vllm/model_executor/layers/quantization/utils/quant_utils.py:scaled_quantize.
        vllm_fp8.init_fp8_linear_kernel = lambda *args, **kwargs: None
        super().__init__(quant_config)
        self.use_marlin = True

        self.linear_config = linear_config
        if self.linear_config.enable_quantized_matmul_kernel and not self.linear_config.requant_block_size:
            raise ValueError(
                "You should set REQUANTIZE_BLOCK_SIZE to enable quantized matmul kernel. Please set the value or disable the quantized matmul kernel."
            )
        if not self.linear_config.enable_quantized_matmul_kernel and self.linear_config.requant_block_size:
            raise ValueError(
                "Blockwise quantization is supported by quantized matmul kernel. Please enable quantized_matmul_kernel or unset the quantize block size to trigger XLA per-channel quantization."
            )

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert isinstance(layer, vllm_linear.LinearBase)

        assert self.block_quant
        weight = t2j(layer.weight, use_dlpack=False)
        delattr(layer, "weight")

        weight_scale = t2j(layer.weight_scale_inv, use_dlpack=False)
        delattr(layer, "weight_scale_inv")

        if layer.bias is not None and not layer.skip_bias_add:
            if layer.return_bias:
                logger.warning_once("Bias might return incorrect value.")
            bias = t2j(layer.bias, use_dlpack=False)
            delattr(layer, "bias")
        else:
            bias = None

        weights = common_fp8.process_blockwise_fp8_linear_weights(
            weight,
            weight_scale,
            bias=bias,
            weight_block_size=tuple(self.weight_block_size),
            requant_block_size=self.linear_config.requant_block_size,
            output_sizes=tuple(self.linear_config.output_sizes),
            requant_weight_dtype=self.linear_config.requant_weight_dtype,
            fuse_matmuls=self.linear_config.fuse_matmuls,
            n_shards=self.linear_config.n_shards)
        if self.linear_config.enable_quantized_matmul_kernel:
            # The quantized_matmul_kernel expects weight scales shaped (n_out_features, 1, n_blocks) for blockwisze quantization.
            weights.weight_scale = jnp.expand_dims(
                jnp.transpose(weights.weight_scale),
                axis=1,
            )
        weights = torch_view(
            shard_linear_weights(
                weights,
                mesh=self.linear_config.mesh,
                weight_p_spec=self.linear_config.weight_sharding,
                bias_p_spec=self.linear_config.bias_sharding,
            ))

        if self.linear_config.fuse_matmuls:
            layer.weight = Parameter(weights.weight, requires_grad=False)
            layer.weight_scale = Parameter(weights.weight_scale,
                                           requires_grad=False)
            if bias is not None:
                layer.bias = Parameter(weights.bias, requires_grad=False)
        else:
            layer.weight = to_parameter_list(weights.weight)
            layer.weight_scale = to_parameter_list(weights.weight_scale)
            if bias is not None:
                layer.bias = to_parameter_list(weights.bias)

    def apply(self,
              layer: torch.nn.Module,
              x: torch.Tensor,
              bias: Optional[torch.Tensor] = None) -> torch.Tensor:
        with jax.named_scope(layer._get_name()):
            x_jax = jax_view(x)
            bias_jax = jax_view(
                bias) if bias is not None and not layer.skip_bias_add else None
            if self.linear_config.fuse_matmuls:
                weight_jax = jax_view(layer.weight)
                weight_scale_jax = jax_view(layer.weight_scale)
                out = self._apply_fused(x_jax, weight_jax, weight_scale_jax,
                                        bias_jax)
            else:
                assert isinstance(layer.weight, torch.nn.ParameterList)
                assert isinstance(layer.weight_scale, torch.nn.ParameterList)
                # jax_view cannot handle ParameterList directly, so we explicitly
                # convert them to list of jax.Array.
                weight_and_scale = [
                    (jax_view(w), jax_view(s))
                    for w, s in zip(layer.weight, layer.weight_scale)
                ]
                if bias is not None and not layer.skip_bias_add:
                    assert isinstance(bias, torch.nn.ParameterList)
                    bias_jax = [jax_view(b) for b in bias]
                out = self._apply_split(x_jax,
                                        weight_and_scale,
                                        bias_jax,
                                        mesh=self.linear_config.mesh)
            return torch_view(out)


class VllmFp8MoEMethod(vllm_fp8.Fp8MoEMethod):

    def __init__(self,
                 quant_config: vllm_fp8.Fp8Config,
                 layer: torch.nn.Module,
                 mesh: Mesh,
                 ep_axis_name: str = "model"):
        FusedMoEMethodBase.__init__(self, layer.moe_config)
        self.quant_config = quant_config
        self.weight_block_size = self.quant_config.weight_block_size
        self.block_quant: bool = self.weight_block_size is not None
        self.weight_scale_name = ("weight_scale_inv"
                                  if self.block_quant else "weight_scale")
        self.fp8_backend = None

        self.mesh = mesh
        self.moe_backend = select_moe_backend_from_fused_moe_config(self.moe)

        self.extra_backend_kwargs = {}
        if self.moe_backend == MoEBackend.FUSED_MOE:
            self.extra_backend_kwargs = dict(ep_axis_name=ep_axis_name, )

    @property
    def is_monolithic(self) -> bool:
        return True

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert isinstance(layer, FusedMoE)

        assert self.block_quant
        assert not self.moe.has_bias

        w13_weight = t2j(layer.w13_weight, use_dlpack=False)
        w13_weight_scale = t2j(layer.w13_weight_scale_inv, use_dlpack=False)

        w2_weight = t2j(layer.w2_weight, use_dlpack=False)
        w2_weight_scale = t2j(layer.w2_weight_scale_inv, use_dlpack=False)

        # TODO: do we need to support bias?
        input_weights = FusedMoEWeights(
            w13_weight=w13_weight,
            w13_weight_scale=w13_weight_scale,
            w13_bias=None,
            w2_weight=w2_weight,
            w2_weight_scale=w2_weight_scale,
            w2_bias=None,
        )

        weight_block_size = None
        if self.weight_block_size is not None:
            weight_block_size = tuple(self.weight_block_size)

        weights = process_fp8_moe_weights(
            input_weights,
            moe_backend=self.moe_backend,
            mesh=self.mesh,
            activation=layer.activation.value,
            # Convert to tuple so jax jit can hash it
            weight_block_size=weight_block_size,
        )
        weights = torch_view(
            shard_moe_weights(weights, self.moe_backend, self.mesh))

        layer.w13_weight = Parameter(weights.w13_weight, requires_grad=False)
        layer.w2_weight = Parameter(weights.w2_weight, requires_grad=False)

        layer.w13_weight_scale_inv = Parameter(weights.w13_weight_scale,
                                               requires_grad=False)
        layer.w2_weight_scale_inv = Parameter(weights.w2_weight_scale,
                                              requires_grad=False)

    def apply_monolithic(
        self,
        layer: FusedMoE,
        x: torch.Tensor,
        router_logits: torch.Tensor,
    ) -> torch.Tensor:

        weights = FusedMoEWeights(
            w13_weight=jax_view(layer.w13_weight),
            w13_weight_scale=jax_view(layer.w13_weight_scale_inv),
            w13_bias=jax_view(layer.w13_bias) if self.moe.has_bias else None,
            w2_weight=jax_view(layer.w2_weight),
            w2_weight_scale=jax_view(layer.w2_weight_scale_inv),
            w2_bias=jax_view(layer.w2_bias) if self.moe.has_bias else None,
        )
        return vllm_moe_apply(layer=layer,
                              weights=weights,
                              quant_method_instance=self,
                              x=x,
                              router_logits=router_logits)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

import jax
import jax.numpy as jnp

# Adjust the import path to match your project structure
# yapf: disable
from tpu_inference import envs
from tpu_inference.layers.jax.moe.utils import (MoEBackend,
                                                get_all_to_all_params_fn,
                                                global_permute_fn, gmm_fn,
                                                local_permute_fn,
                                                select_moe_backend,
                                                sort_activations_fn,
                                                unpermute_fn)

# yapf: enable


class TestMoEUtils(unittest.TestCase):

    def setUp(self):
        self.key = jax.random.PRNGKey(0)

    def test_sort_activations_fn(self):
        """Test stateless sorting of activations."""
        inputs = jnp.array([[10], [20], [30], [40]])
        indices = jnp.array([3, 0, 1, 2])
        sorted_out = sort_activations_fn(inputs, indices)
        expected = jnp.array([[40], [10], [20], [30]])
        self.assertTrue(jnp.array_equal(sorted_out, expected))

    def test_global_permute_fn(self):
        """Test global permutation logic."""
        # Setup: 2 tokens, embedding dim 1.
        # Expert assignments: Token 0 -> Expert 1, Token 1 -> Expert 0.
        # num_experts_per_tok = 1, num_local_experts = 2.
        inputs = jnp.array([[10.0], [20.0]])
        selected_experts = jnp.array([[1], [0]])

        # Expected behavior:
        # Flattened experts: [1, 0]
        # Sorted indices should bring expert 0 first, then expert 1.
        # Indices: [1, 0] (since expert 0 is at index 1, expert 1 is at index 0)

        sorted_inputs, sort_indices, group_sizes, expert_assignments = \
            global_permute_fn(inputs, selected_experts,
                              num_experts_per_tok=1,
                              num_local_experts=2)

        # Check group sizes (1 token for expert 0, 1 token for expert 1)
        self.assertTrue(jnp.array_equal(group_sizes, jnp.array([1, 1])))

        # Check sorted inputs: Expert 0's token (20.0) first, then Expert 1's (10.0)
        expected_inputs = jnp.array([[20.0], [10.0]])
        self.assertTrue(jnp.array_equal(sorted_inputs, expected_inputs))

        # Check sort indices
        expected_indices = jnp.array([1, 0])
        self.assertTrue(jnp.array_equal(sort_indices, expected_indices))

    def test_unpermute_fn(self):
        """Test unpermute logic (recombining sorted expert outputs)."""
        # 2 tokens, 2 experts per token. Flattened = 4 computations.
        # Let's say inputs were originally T0, T1.
        # Sorted order processed was: T0_e1, T1_e0, T0_e2, T1_e2 (just hypothetical).
        # We test that it puts them back into [T0_combined, T1_combined].

        num_tokens = 2
        k = 2
        dim = 4

        # Processed tokens (T*K, D)
        processed = jnp.ones((num_tokens * k, dim))

        # Indices that were used to sort them.
        # If we provide identity indices, unpermute should just reshape and sum weighted.
        sort_indices = jnp.arange(num_tokens * k)

        # Router weights: (T, K)
        weights = jnp.full((num_tokens, k), 0.5)

        output = unpermute_fn(processed,
                              sort_indices,
                              weights,
                              num_experts_per_tok=k,
                              output_dtype=jnp.float32)

        # Expected: sum(token_val * 0.5) over k=2 -> 1.0 * 0.5 + 1.0 * 0.5 = 1.0
        self.assertEqual(output.shape, (num_tokens, dim))
        self.assertTrue(jnp.allclose(output, jnp.ones((num_tokens, dim))))

    def test_local_permute_fn(self):
        """Test local permutation slicing."""
        # 4 tokens total across all experts
        # Global Group Sizes: [[1, 1, 1, 1]] (1 token per expert, 4 experts)
        # Local expert size: 2.
        # Shard 0 owns experts [0, 1]. Shard 1 owns experts [2, 3].

        inputs = jnp.arange(4)[:, None]  # [0, 1, 2, 3]
        global_group_sizes = jnp.array([[1, 1, 1, 1]])

        # Test Shard 0 (Experts 0, 1)
        # It should slice the first 2 counts from group sizes -> [1, 1]. Total 2.
        # is_offset=False (Token sharded usually implies standard modulo or repetition,
        # but here we test the logic flow).

        # Mocking logic: effectively testing the dynamic slice and argsort
        # For this test, we construct inputs to match the logic.

        _, _, size, expert_ids = local_permute_fn(inputs,
                                                  global_group_sizes,
                                                  local_expert_size=2,
                                                  shard_index=0,
                                                  is_offset=False)

        # We expect it to process the local experts
        self.assertEqual(list(size), [1, 1])
        self.assertTrue(jnp.array_equal(expert_ids, jnp.array([0, 1, 1, 1])))

    def test_get_all_to_all_params_fn(self):
        """Test parameter generation for all-to-all communication."""
        # 2 Shards.
        # Group sizes matrix (conceptually):
        # Shard 0 sends 10 to Shard 0, 20 to Shard 1
        # Shard 1 sends 30 to Shard 0, 40 to Shard 1
        # Passed as flat array or specific shape depending on usage,
        # but the function expects array access logic.

        # Let's assume input is (2, 2) for (senders, receivers)
        group_sizes = jnp.array([[10, 20], [30, 40]])

        # Test Shard 0 params
        shard_id = 0

        # Strategy: INPUT_OFFSET (cumsum of what I send)
        # I am Shard 0. I send [10, 20]. offsets: [0, 10]
        input_offsets, send_sizes, output_offsets, recv_sizes = \
            get_all_to_all_params_fn(group_sizes, shard_id,
                                     num_expert_parallelism=2,
                                     is_batch_sharded=True)

        self.assertTrue(jnp.array_equal(input_offsets, jnp.array([0, 10])))
        self.assertTrue(jnp.array_equal(send_sizes, jnp.array([10, 20])))

        # RECV_SIZE: I am Shard 0. I receive column 0: [10, 30]
        self.assertTrue(jnp.array_equal(recv_sizes, jnp.array([10, 30])))

    @mock.patch('tpu_inference.layers.jax.moe.utils.megablox_gmm')
    @mock.patch(
        'tpu_inference.layers.jax.moe.utils.round_up_to_multiple_of_128_within_limit'
    )
    def test_gmm_fn_megablox(self, mock_round, mock_megablox):
        """Test GMM function calls MegaBlocks correctly."""
        inputs = jnp.ones((128, 64))
        kernel = jnp.ones((1, 64, 128))  # [E, K, N]
        group_sizes = jnp.array([128])
        tile_size = (128, 128, 128)

        mock_megablox.return_value = jnp.zeros((128, 128))
        mock_round.side_effect = lambda x, y: x

        gmm_fn(inputs,
               kernel,
               group_sizes,
               tile_size,
               moe_backend=MoEBackend.MEGABLX_GMM,
               dtype=jnp.bfloat16,
               quantized_dtype=None)

        mock_megablox.assert_called_once()

    @mock.patch('tpu_inference.layers.jax.moe.utils.megablox_gmm')
    @mock.patch(
        'tpu_inference.layers.jax.moe.utils.round_up_to_multiple_of_128_within_limit'
    )
    def test_gmm_fn_megablox_quantized(self, mock_round, mock_megablox):
        """Test GMM function calls MegaBlocks correctly."""
        inputs = jnp.ones((128, 64))
        kernel = jnp.ones((1, 64, 128)), jnp.ones((1, 1, 128))
        group_sizes = jnp.array([128])
        tile_size = (128, 128, 128)

        mock_megablox.return_value = jnp.zeros((128, 128))
        mock_round.side_effect = lambda x, y: x

        gmm_fn(inputs,
               kernel,
               group_sizes,
               tile_size,
               moe_backend=MoEBackend.MEGABLX_GMM,
               dtype=jnp.bfloat16,
               quantized_dtype=jnp.float8_e4m3fn)

        mock_megablox.assert_called_once()


class TestMoESelector(unittest.TestCase):

    def setUp(self):
        self.key = jax.random.PRNGKey(0)

    def test_select_moe_backend_defaults(self):
        """Test default backend selection (GMM_TP/GMM_EP)."""
        # Ensure all explicit backend flags are False
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', False), \
             mock.patch.object(envs, 'USE_UNFUSED_MEGABLOCKS', False):

            # Case 1: No EP enabled -> Fallback to GMM_TP
            backend_tp = select_moe_backend(use_ep=False)
            self.assertEqual(backend_tp, MoEBackend.GMM_TP)

            # Case 2: EP enabled -> Fallback to GMM_EP
            backend_ep = select_moe_backend(use_ep=True)
            self.assertEqual(backend_ep, MoEBackend.GMM_EP)

    def test_select_moe_backend_priority(self):
        """Test priority logic for backend selection."""

        # Case 1: Fused MoE (Highest priority when USE_MOE_EP_KERNEL=True AND use_ep=True)
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', True), \
             mock.patch.object(envs, 'USE_UNFUSED_MEGABLOCKS', False):
            self.assertEqual(select_moe_backend(use_ep=True),
                             MoEBackend.FUSED_MOE)

        # Case 1.5: Fused MoE Flag is True, but use_ep is False.
        # Should fall through to defaults (GMM_TP) because Fused kernel requires EP.
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', True), \
             mock.patch.object(envs, 'USE_UNFUSED_MEGABLOCKS', False):
            self.assertEqual(select_moe_backend(use_ep=False),
                             MoEBackend.GMM_TP)

        # Case 2: MegaBlocks (Next priority)
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', False), \
             mock.patch.object(envs, 'USE_UNFUSED_MEGABLOCKS', True):
            self.assertEqual(select_moe_backend(use_ep=True),
                             MoEBackend.MEGABLX_GMM)
            self.assertEqual(select_moe_backend(use_ep=False),
                             MoEBackend.MEGABLX_GMM)
        # Case 2: Dense (Next priority)
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', False), \
             mock.patch.object(envs, 'USE_DENSE_MOE', True):
            self.assertEqual(select_moe_backend(use_ep=True),
                             MoEBackend.GMM_EP)
            self.assertEqual(select_moe_backend(use_ep=False),
                             MoEBackend.DENSE_MAT)

        # Default: GMM_TP
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', False), \
             mock.patch.object(envs, 'USE_DENSE_MOE', False):
            self.assertEqual(select_moe_backend(use_ep=True),
                             MoEBackend.GMM_EP)
            self.assertEqual(select_moe_backend(use_ep=False),
                             MoEBackend.GMM_TP)

    def test_select_moe_backend_precedence_conflict(self):
        """Test precedence when multiple flags are enabled."""
        # EP Kernel + MegaBlocks: EP Kernel takes precedence if use_ep=True
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', True), \
             mock.patch.object(envs, 'USE_UNFUSED_MEGABLOCKS', True):
            self.assertEqual(select_moe_backend(use_ep=True),
                             MoEBackend.FUSED_MOE)

        # MegaBlocks + Ragged Dot: MegaBlocks takes precedence
        with mock.patch.object(envs, 'USE_MOE_EP_KERNEL', False), \
             mock.patch.object(envs, 'USE_UNFUSED_MEGABLOCKS', True):
            self.assertEqual(select_moe_backend(use_ep=True),
                             MoEBackend.MEGABLX_GMM)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import InitVar, dataclass
from typing import Any, NamedTuple, Optional

import jax
import jax.numpy as jnp
from flax import nnx
from jax import lax
from jax.sharding import Mesh, NamedSharding
from jax.sharding import PartitionSpec as P
from jax.sharding import Sharding

from tpu_inference.layers.common.attention_interface import \
    sharded_flash_attention
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.common.quantization import quantize_kv
from tpu_inference.layers.jax.attention.attention import Attention, KVCache
from tpu_inference.layers.jax.base import create_param
from tpu_inference.layers.jax.rope_interface import apply_rope
from tpu_inference.logger import init_logger

logger = init_logger(__name__)


class L2Norm(nnx.Module):
    """
  Implementation of L2 Norm in JAX (taken from MaxText repo - maxtext/MaxText/layers/attentions.py).

  Attributes:
    eps: float, epsilon used for numerical stability (default value should be ok for most cases).
  """

    def __init__(self, eps: float = 1e-6):
        self.eps = eps

    def __call__(self, x):
        return x * jax.lax.rsqrt(
            jnp.mean(x**2, axis=-1, keepdims=True) + self.eps)


@dataclass(kw_only=True)
class Llama4Attention(Attention):
    use_qk_norm: bool
    temperature_tuning: bool
    temperature_tuning_floor_scale: float
    temperature_tuning_scale: float
    activation_attention_td: Sharding
    activation_attention_out_td: Sharding

    def __call__(self,
                 x,
                 is_prefill,
                 kv_cache: KVCache,
                 attention_metadata: AttentionMetadata,
                 use_attention_rope: bool = True,
                 **kwargs):
        """Performs the forward pass of the attention module.

        This method computes the attention output by projecting the input `x`
        to queries, keys, and values, applying RoPE and L2Norm if specified,
        performing scaled dot-product attention, and projecting the results
        back to the model dimension.
        If no RoPE (NoPE) is specified, one can also perform temperature tuning
        which is useful to combat dilution of attention scores in long-context attention.

        Args:
            x: The input tensor of shape `(seq_len, d_model)`.
            is_prefill: Whether the operation mode is prefill (otherwise it is generate).
            kv_cache: The key-value cache for storing past attention states.
            attention_metadata: Metadata for attention, such as input positions.
            use_attention_rope: Whether to use RoPE.

        Returns:
            A tuple containing:
                - The updated KV cache.
                - The attention output tensor of shape
                  `(batch_size, seq_len, d_model)`.
        """
        md = attention_metadata
        x = jnp.asarray(x, self.dtype)
        x_SD = lax.with_sharding_constraint(x, self.activation_attention_td)
        x_q_TD = lax.with_sharding_constraint(x, self.activation_q_td)
        rope_scaling = self.rope_scaling
        rope_theta = self.rope_theta
        H = self.head_dim
        l2_norm = L2Norm()

        with jax.named_scope("q_proj"):
            q_TNH = jnp.einsum('TD,DNH -> TNH', x_q_TD,
                               self.kernel_q_proj_DNH.value)
            if use_attention_rope:
                q_TNH = apply_rope(q_TNH, md.input_positions, H, rope_theta,
                                   rope_scaling, self.rope_input_ordering)

                # Apply normaliation after RoPE
                if self.use_qk_norm:
                    q_TNH = l2_norm(q_TNH)
            else:
                if self.temperature_tuning:
                    q_TNH = self.apply_temperature_tuning(md, q_TNH)

            q_TNH = lax.with_sharding_constraint(
                q_TNH, NamedSharding(self.mesh, self.query_tnh))
        with jax.named_scope("k_proj"):
            k_SKH = jnp.einsum('SD,DKH -> SKH', x_SD,
                               self.kernel_k_proj_DKH.value)
            if use_attention_rope:
                k_SKH = apply_rope(k_SKH, md.input_positions, H, rope_theta,
                                   rope_scaling, self.rope_input_ordering)

                # Apply normaliation after RoPE
                if self.use_qk_norm:
                    k_SKH = l2_norm(k_SKH)
            k_SKH = lax.with_sharding_constraint(
                k_SKH, NamedSharding(self.mesh, self.keyvalue_skh))

        with jax.named_scope("v_proj"):
            v_SKH = jnp.einsum('SD,DKH -> SKH', x_SD,
                               self.kernel_v_proj_DKH.value)
            v_SKH = lax.with_sharding_constraint(
                v_SKH, NamedSharding(self.mesh, self.keyvalue_skh))

        q_scale = k_scale = v_scale = None
        if self.kv_cache_quantized_dtype:
            # TODO(kyuyeunk/jacobplatin): Enable w8a8 when VREG spill issue is resolved.
            k_scale = self._k_scale
            v_scale = self._v_scale
            k_SKH, v_SKH = quantize_kv(self.kv_cache_quantized_dtype, k_SKH,
                                       v_SKH, k_scale, v_scale)
        with jax.named_scope("attn_op"):
            new_kv_cache, outputs_TNH = self.attention(is_prefill,
                                                       kv_cache,
                                                       q_TNH,
                                                       k_SKH,
                                                       v_SKH,
                                                       attention_metadata,
                                                       self.mesh,
                                                       q_scale=q_scale,
                                                       k_scale=k_scale,
                                                       v_scale=v_scale,
                                                       **kwargs)
        with jax.named_scope("o_proj"):
            o_TD = jnp.einsum('TNH,NHD -> TD', outputs_TNH,
                              self.kernel_o_proj_NHD.value)
            o_TD = lax.with_sharding_constraint(
                o_TD, self.activation_attention_out_td)
        return new_kv_cache, o_TD

    def apply_temperature_tuning(self, md: AttentionMetadata,
                                 input_arr_TNH: jax.Array) -> jax.Array:
        """Applies temperature tuning to the input array of shape (T, N, H).
        Args:
            md: AttentionMetadata object containing the input positions.
            input_arr_TNH: Input array of shape (T, N, H) which will have scaled temperatures applied.
        """
        attn_scales = (jnp.log(
            jnp.floor((md.input_positions.astype(self.dtype) + 1.0) /
                      self.temperature_tuning_floor_scale) + 1.0) *
                       self.temperature_tuning_scale + 1.0)
        return input_arr_TNH * attn_scales[:, None, None]


class SegmentIds(NamedTuple):
    """SegmentIds for Q and KV sequences.

  SegmentIds are used to generate segment mask, which prevents attention between
  different segments in the input sequence. Each array is a list of ids
  (integers).
  Only the token with the same id can attend to each other.

  Attributes:
    q: segment ids along the Q sequence.
    kv: segment ids along the KV sequence.
  """

    q: jax.Array  # [batch_size, q_seq_len]
    kv: jax.Array  # [batch_size, kv_seq_len]


@dataclass(kw_only=True)
class Llama4VisionAttention(nnx.Module):
    hidden_size: int
    num_attention_heads: int
    num_key_value_heads: int
    head_dim: int
    rope_theta: float
    rope_scaling: Optional[dict[str, Any]]
    dtype: jnp.dtype
    mesh: Mesh
    use_qk_norm: bool
    temperature_tuning: bool
    temperature_tuning_floor_scale: float
    temperature_tuning_scale: float
    activation_attention_td: Sharding
    activation_attention_out_td: Sharding
    is_causal: bool = False
    kv_cache_quantized_dtype: Optional[jnp.dtype] = None
    rngs: InitVar[nnx.Rngs]

    dnh_sharding: Sharding = ()
    dkh_sharding: Sharding = ()
    nhd_sharding: Sharding = ()
    activation_q_td: Sharding = ()
    query_tnh: P = P()
    keyvalue_skh: P = P()
    rope_input_ordering: str = "interleaved"  # Vision config default

    _q_scale: float = 1.0
    _k_scale: float = 1.0
    _v_scale: float = 1.0

    def __post_init__(self, rngs: nnx.Rngs):
        """Initializes the weight kernels for Q, K, V, and O projections."""
        N = self.num_attention_heads
        K = self.num_key_value_heads
        D = self.hidden_size
        H = self.head_dim
        random_init = False

        self.kernel_q_proj_DNH = create_param(rngs, (D, N, H),
                                              self.dnh_sharding,
                                              self.dtype,
                                              random_init=random_init)
        self.kernel_k_proj_DKH = create_param(rngs, (D, K, H),
                                              self.dkh_sharding,
                                              self.dtype,
                                              random_init=random_init)
        self.kernel_v_proj_DKH = create_param(rngs, (D, K, H),
                                              self.dkh_sharding,
                                              self.dtype,
                                              random_init=random_init)
        self.kernel_o_proj_NHD = create_param(rngs, (N, H, D),
                                              self.nhd_sharding,
                                              self.dtype,
                                              random_init=random_init)

        self.bias_q_proj_NH = create_param(rngs, (N, H),
                                           self.nhd_sharding,
                                           self.dtype,
                                           random_init=random_init)
        self.bias_k_proj_KH = create_param(rngs, (K, H),
                                           self.dnh_sharding,
                                           self.dtype,
                                           random_init=random_init)
        self.bias_v_proj_KH = create_param(rngs, (K, H),
                                           self.dkh_sharding,
                                           self.dtype,
                                           random_init=random_init)
        self.bias_o_proj_D = create_param(rngs, (D, ),
                                          self.dkh_sharding,
                                          self.dtype,
                                          random_init=random_init)

    def __call__(self,
                 x,
                 is_prefill,
                 kv_cache: KVCache,
                 attention_metadata: AttentionMetadata,
                 freqs_cis: jax.Array,
                 use_attention_rope: bool = True,
                 **kwargs):

        # md = attention_metadata
        x = jnp.asarray(x, self.dtype)
        x_SD = lax.with_sharding_constraint(x, self.activation_attention_td)
        x_q_TD = lax.with_sharding_constraint(x, self.activation_q_td)

        rope_scaling = self.rope_scaling
        rope_theta = self.rope_theta
        H = self.head_dim

        with jax.named_scope("q_proj"):
            q_TNH = jnp.einsum('TD,DNH -> TNH', x_q_TD,
                               self.kernel_q_proj_DNH.value)
            q_TNH += self.bias_q_proj_NH.value[None, ...]

            if use_attention_rope:
                q_TNH = apply_rope(q_TNH, freqs_cis, H, rope_theta,
                                   rope_scaling, self.rope_input_ordering)
            q_TNH = lax.with_sharding_constraint(q_TNH, self.query_tnh)

        with jax.named_scope("k_proj"):
            k_SKH = jnp.einsum('SD,DKH -> SKH', x_SD,
                               self.kernel_k_proj_DKH.value)
            k_SKH += self.bias_k_proj_KH.value[None, ...]

            if use_attention_rope:
                k_SKH = apply_rope(k_SKH, freqs_cis, H, rope_theta,
                                   rope_scaling, self.rope_input_ordering)
            k_SKH = lax.with_sharding_constraint(k_SKH, self.keyvalue_skh)

        with jax.named_scope("v_proj"):
            v_SKH = jnp.einsum('SD,DKH -> SKH', x_SD,
                               self.kernel_v_proj_DKH.value)
            v_SKH += self.bias_v_proj_KH.value[None, ...]
            v_SKH = lax.with_sharding_constraint(v_SKH, self.keyvalue_skh)

        if self.kv_cache_quantized_dtype:
            k_scale = self._k_scale
            v_scale = self._v_scale
            k_SKH, v_SKH = quantize_kv(self.kv_cache_quantized_dtype, k_SKH,
                                       v_SKH, k_scale, v_scale)

        T_attn, N, H = q_TNH.shape
        B = 1
        BLOCK_SIZE = 128
        pad_len = (BLOCK_SIZE - (T_attn % BLOCK_SIZE)) % BLOCK_SIZE

        # Pad Q/K/V
        q_TNH = jnp.pad(q_TNH, [(0, pad_len), (0, 0), (0, 0)],
                        constant_values=0)
        q_TNH = jnp.expand_dims(q_TNH, axis=0)
        k_SKH = jnp.pad(k_SKH, [(0, pad_len), (0, 0), (0, 0)],
                        constant_values=0)
        k_SKH = jnp.expand_dims(k_SKH, axis=0)
        v_SKH = jnp.pad(v_SKH, [(0, pad_len), (0, 0), (0, 0)],
                        constant_values=0)
        v_SKH = jnp.expand_dims(v_SKH, axis=0)

        q_BNTH = jnp.transpose(q_TNH, (0, 2, 1, 3))
        k_BKTH = jnp.transpose(k_SKH, (0, 2, 1, 3))
        v_BKTH = jnp.transpose(v_SKH, (0, 2, 1, 3))

        # Mask Padding
        valid_ids = jnp.zeros((B, T_attn), dtype=jnp.int32)
        pad_ids = jnp.ones((B, pad_len), dtype=jnp.int32)
        segment_ids_q = jnp.concatenate([valid_ids, pad_ids], axis=1)

        segment_ids = SegmentIds(q=segment_ids_q, kv=segment_ids_q)

        with jax.named_scope("flash_attn_op"):
            outputs_BNTH = sharded_flash_attention(
                mesh=self.mesh,
                causal=self.is_causal,
                sm_scale=self.head_dim**-0.5,
            )(q_BNTH, k_BKTH, v_BKTH, segment_ids)
            new_kv_cache = kv_cache

        outputs_TBH = jnp.transpose(outputs_BNTH, (2, 0, 1, 3))
        outputs_TBH = outputs_TBH[:T_attn, ...]
        outputs_TNH = jnp.squeeze(outputs_TBH, axis=1)

        with jax.named_scope("o_proj"):
            o_TD = jnp.einsum('TNH,NHD -> TD', outputs_TNH,
                              self.kernel_o_proj_NHD.value)
            o_TD += self.bias_o_proj_D.value
            o_TD = lax.with_sharding_constraint(
                o_TD, self.activation_attention_out_td)

        return new_kv_cache, o_TD


# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
from unittest.mock import MagicMock

import jax
import numpy as np
import torch
from flax import nnx
from jax.sharding import Mesh
from safetensors.torch import save_file
from torch import nn
from vllm.config import set_current_vllm_config
from vllm.model_executor.model_loader import LoadConfig, get_model_loader

from tpu_inference.layers.jax import JaxModule
from tpu_inference.layers.jax.linear import JaxLinear
from tpu_inference.models.jax.utils.weight_utils import LoadableWithIterator


class TorchMLP(nn.Module):
    """MLP implemented with PyTorch."""

    def __init__(self):
        super().__init__()
        self.w1 = nn.Linear(2, 6)
        self.act = nn.ReLU()
        self.w2 = nn.Linear(6, 2, bias=True)

    def forward(self, x):
        x = self.w1(x)
        x = self.act(x)
        x = self.w2(x)
        return x


class JaxMLP(JaxModule, LoadableWithIterator):
    """MLP implemented with JAX."""

    def __init__(self, rngs):
        super().__init__()
        self.w1 = JaxLinear(2, 6, rngs)
        self.act = nnx.relu
        self.w2 = JaxLinear(6, 2, rngs, use_bias=True)

    def __call__(self, x):
        x = self.w1(x)
        x = self.act(x)
        x = self.w2(x)
        return x


class TestJaxAutoWeightsLoader:

    def test_load_from_safetensors(self):
        """Load weights from a safetensors file saved from a PyTorch model.
        """
        torch_model = TorchMLP()
        with torch.no_grad():
            torch_model.w1.weight.fill_(1.1)
            torch_model.w2.weight.fill_(0.9)
            torch_model.w2.bias.fill_(0.1)

        mock_vllm_config = MagicMock()
        mock_vllm_config.parallel_config = MagicMock()
        mock_vllm_config.parallel_config.enable_expert_parallel = False

        # Save the PyTorch model weights to a safetensors file. Load them
        # into the JAX model.
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_file_path = os.path.join(tmpdir, "model.safetensors")
            save_file(torch_model.state_dict(), tmp_file_path)

            devices = jax.local_devices()
            mesh = Mesh(devices, axis_names=('p', ))
            with jax.set_mesh(mesh), set_current_vllm_config(mock_vllm_config):
                jax_model = JaxMLP(rngs=nnx.Rngs(0))

                model_config = MagicMock()
                model_config.quantization = None
                model_config.model = tmpdir
                model_config.revision = None

                loader = get_model_loader(
                    LoadConfig(load_format="safetensors"))
                loader.load_weights(jax_model, model_config)

        np.testing.assert_allclose(torch_model.w1.weight.T.detach().numpy(),
                                   jax_model.w1.weight.value)
        np.testing.assert_allclose(torch_model.w2.weight.T.detach().numpy(),
                                   jax_model.w2.weight.value)

        # Forward pass to verify correctness.
        input_values = [[0.1, 0.2], [0.3, 0.4]]
        torch_input = torch.tensor(input_values)
        jax_input = np.array(input_values)
        torch_output = torch_model(torch_input).detach().numpy()
        jax_output = jax_model(jax_input)
        np.testing.assert_allclose(torch_output,
                                   jax_output,
                                   rtol=1e-3,
                                   atol=1e-2)
